/*
 *  Copyright (c) xlightweb.org, 2008 - 2009. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;

import java.io.IOException;
import java.net.InetAddress;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.text.SimpleDateFormat;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.Executor;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicReference;
import java.util.logging.Level;
import java.util.logging.Logger;


import org.xlightweb.HttpUtils.HttpConnectionHandlerInfo;
import org.xsocket.DataConverter;
import org.xsocket.Execution;
import org.xsocket.IDestroyable;
import org.xsocket.MaxReadSizeExceededException;
import org.xsocket.SerializedTaskQueue;
import org.xsocket.connection.ConnectionUtils;
import org.xsocket.connection.IConnectHandler;
import org.xsocket.connection.IConnectionTimeoutHandler;
import org.xsocket.connection.IDataHandler;
import org.xsocket.connection.IDisconnectHandler;
import org.xsocket.connection.IIdleTimeoutHandler;
import org.xsocket.connection.INonBlockingConnection;
import org.xsocket.connection.IWriteCompletionHandler;
import org.xsocket.connection.NonBlockingConnectionPool;




/**
 * 
 * Implementation base for the {@link HttpClientConnection} and {@link HttpServerConnection} class
 * 
 * <br/><br/><b>This is a xSocket internal class and subject to change</b> 
 * 
 * @author grro@xlightweb.org
 */
public abstract class AbstractHttpConnection implements IHttpConnection, IDestroyable {
	

	private static final Logger LOG = Logger.getLogger(AbstractHttpConnection.class.getName());

	
	private static final NullMessageWriter NULL_BODYWRITER = new NullMessageWriter(); 
	
	
	// max write buffer size
	private static final String DEFAULT_MAX_WRITE_BUFFER_SIZE = "65536";
	private static final int MAX_WRITEBUFFER_SIZE = Integer.parseInt(System.getProperty("org.xlightweb.connection.max_write_buffer_size", DEFAULT_MAX_WRITE_BUFFER_SIZE));
	
		
	// timer support 
	private static final Timer TIMER = new Timer("xHttpTimer", true);

		
	// underlying tcp connection
	private final INonBlockingConnection tcpConnection;

	// http connection properties
	private final AtomicBoolean isPersistent = new AtomicBoolean(true);


	// life cycle management (output body close listener support)
	private final AtomicReference<BodyDataSink> bodyDataSink = new AtomicReference<BodyDataSink>();

	
	
	// attachment management
	private Object attachment;
	

	// protocol handling
	private final DataHandler dataHandler = new DataHandler();
	private final AbstractHttpProtocolHandler protocolHandler;
	
	// connection life cycle handler support 
	private final Set<IHttpConnectionHandler> connectionHandlers = Collections.synchronizedSet(new HashSet<IHttpConnectionHandler>());
	private final AtomicBoolean isDisconnected = new AtomicBoolean(false);
	
	
	// call back processing
	private final MultimodeExecutor multimodeExcutor = new MultimodeExecutor();


	// timeout support
	private long bodyDataReceiveTimeoutMillis = Long.MAX_VALUE;
	private long lastTimeDataWritten = System.currentTimeMillis();
	private long lastTimeHeaderReceivedMillis = System.currentTimeMillis();
	private long lastTimeMessageTailReceivedMillis = System.currentTimeMillis();
	private long lastTimeDataReceivedMillis = System.currentTimeMillis();


	// statistics
	private final AtomicInteger countSuspendReceiving = new AtomicInteger(0);
	private final AtomicInteger countResumeReceiving = new AtomicInteger(0);
	private final AtomicInteger countReceivedMessages = new AtomicInteger(0);
	private final AtomicInteger countSentMessages = new AtomicInteger(0);
	private int countSendBytes = 0;
	private int countReceivedBytes = 0;


	
	
	
	/**
	 * constructor 
	 * 
	 * @param nativeConnection        the underlying tcp connection
	 * @param isClientSideConnection  true, if http connection is client-side connection
	 * 
	 * @throws IOException if an exception occurs
	 */
	protected AbstractHttpConnection(INonBlockingConnection tcpConnection, boolean isClientSideConnection) throws IOException {
		this.tcpConnection = tcpConnection;
		
		if (isClientSideConnection) {
			protocolHandler = new HttpProtocolHandlerClientSide();
			
		} else {
			protocolHandler = new HttpProtocolHandlerServerSide();
		}
		
		
		// work around to deactivate xsocket's retry read 
		tcpConnection.setMaxReadBufferThreshold(Integer.MAX_VALUE - 1);
		
		tcpConnection.setAutoflush(false);
		tcpConnection.setAttachment(this);
	}
	


	/**
	 * returns the body type string
	 * 
	 * @param body  the body 
	 * @return the body type string
	 */
	protected static String getBodyType(NonBlockingBodyDataSource body) {
		return body.getBodyType().toString();
	}
	
	/**
	 * initialize this connection
	 * @throws IOException  if an exception occurs
	 */
	protected final void init() throws IOException {
		tcpConnection.setHandler(dataHandler);
		onConnect();
	}


	/**
	 * increments the the number of retrieved messages 
	 * @return the number of the received messages after incrementing it
	 */
	protected final int incCountMessageReceived() {
		return countReceivedMessages.incrementAndGet();
	}
	
	
	/**
	 * returns the number of the received messages  
	 * @return the number of the received messages
	 */
	protected int getCountMessagesReceived() {
		return countReceivedMessages.get();
	}
	
	
	/**
	 * returns the number of the send bytes
	 * @return the number of the send bytes
	 */
	protected int getCountSendBytes() {
		return countSendBytes;
	}
	

	/**
	 * returns the number of the received bytes
	 * @return the number of the received bytes
	 */
	protected int getCountReceivedBytes() {
		return countReceivedBytes;
	}
	
	
	/**
	 * increments the the number of sent messages 
	 * @return the number of the sent messages after incrementing it
	 */
	protected final int incCountMessageSent() {
		return countSentMessages.incrementAndGet();
	}
	

	/**
	 * returns the number of the sent messages  
	 * @return the number of the sent messages
	 */
	protected int getCountMessagesSent() {
		return countSentMessages.get();
	}
	


	/**
	 * returns the number of suspend  
	 * @return the number of suspend
	 */
	protected int getCountSuspendReceiving() {
		return countSuspendReceiving.get();
	}
	
	

	/**
	 * returns the number of resume  
	 * @return the number of resume
	 */
	protected int getCountResumeReceiving() {
		return countResumeReceiving.get();
	}
	
	
	/**
	 * sets the time when header or body data has been received
	 * @param time the time when header or body data has been received
	 */
	protected final void setLastTimeDataReceivedMillis(long time) {
		lastTimeDataReceivedMillis = time;
	}

	
	/**
	 * returns the time when header or body data has been received
	 * @return the time when header or body data has been received
	 */
	protected long getLastTimeDataReceivedMillis() {
		return lastTimeDataReceivedMillis;
	}
	
	
	/**
	 * sets the time when body< data has been received  
	 * @param time  the time when body< data has been received
	 */
	protected final void setLastTimeMessageTailReceivedMillis(long time) {
		lastTimeMessageTailReceivedMillis = time;
	}
	
	
	/**
	 * gets the time when body data has been received 
	 * @return the time when body data has been received
	 */
	protected final long getLastTimeMessageTailReceivedMillis() {
		return lastTimeMessageTailReceivedMillis;
	}

	
	/**
	 * sets the time when the (complete) header has been received    
	 * @param time the time when the (complete) header has been received
	 */
	protected final void setLastTimeHeaderReceivedMillis(long time) {
		lastTimeHeaderReceivedMillis = time;
	}
	
	
	/**
	 * gets the time when the (complete) header has been received 
	 * @return the time when the (complete) header has been received
	 */
	protected final long getLastTimeHeaderReceivedMillis() {
		return lastTimeHeaderReceivedMillis;
	}
	
	

	/**
	 * gets the last time data has been written
	 * @return the last time data has been written
	 */
	protected final long getLastTimeWritten() {
		return lastTimeDataWritten;
	}
	

	
	
	/**
	 * returns the assigned executor 
	 * @return the assigned executor
	 */
	protected final IMultimodeExecutor getExecutor() {
		return multimodeExcutor;
	}


	/**
	 * returns the underlying http connection 
	 * @return the underlying http connection 
	 */
	public INonBlockingConnection getUnderlyingTcpConnection() {
		return tcpConnection;
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final long getBodyDataReceiveTimeoutMillis() {
		return bodyDataReceiveTimeoutMillis;
	}	

	
	/**
	 * {@inheritDoc}
	 */
	public final void addConnectionHandler(IHttpConnectionHandler connectionHandler) {
		if (connectionHandler == null) {
			throw new NullPointerException("conection handler has to be set");
		}
		
		connectionHandlers.add(connectionHandler);
		if (isDisconnected.get()) {
			callOnDisconnect(connectionHandler, HttpUtils.getHttpConnectionHandlerInfo(connectionHandler));
		}
	}

	
	/**
	 * {@inheritDoc}
	 */
	public final void removeConnectionHandler(IHttpConnectionHandler connectionHandler) {
		connectionHandlers.remove(connectionHandler);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final void setBodyDataReceiveTimeoutMillis(long bodyDataReceiveTimeoutMillis) {
		this.bodyDataReceiveTimeoutMillis = bodyDataReceiveTimeoutMillis;
	}


	/**
	 * {@inheritDoc}
	 */
	public final void setAttachment(Object attachment) {
		this.attachment = attachment;
		
	}
	

	/**
	 * {@inheritDoc}
	 */
	public final Object getAttachment() {
		return attachment;
	}

	
	/**
	 * {@inheritDoc}
	 */
	public final long getConnectionTimeoutMillis() {
		return tcpConnection.getConnectionTimeoutMillis();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final void setConnectionTimeoutMillis(long timeoutMillis) {
		tcpConnection.setConnectionTimeoutMillis(timeoutMillis);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final long getRemainingMillisToConnectionTimeout() {
		return tcpConnection.getRemainingMillisToConnectionTimeout();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final long getIdleTimeoutMillis() {
		return tcpConnection.getIdleTimeoutMillis();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final void setIdleTimeoutMillis(long timeoutInMillis) {
		tcpConnection.setIdleTimeoutMillis(timeoutInMillis);
	}

	
	/**
	 * {@inheritDoc}
	 */
	public final long getRemainingMillisToIdleTimeout() {
		return tcpConnection.getRemainingMillisToIdleTimeout();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	@SuppressWarnings("unchecked")
	public final Map<String, Class> getOptions() {
		return tcpConnection.getOptions();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final Object getOption(String name) throws IOException {
		return tcpConnection.getOption(name);
	}

	
	/**
	 * {@inheritDoc}
	 */
	public final void setOption(String name, Object value) throws IOException {
		tcpConnection.setOption(name, value);
	}
	

	/**
	 * {@inheritDoc}
	 */
	public final InetAddress getLocalAddress() {
		return tcpConnection.getLocalAddress();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final int getLocalPort() {
		return tcpConnection.getLocalPort();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public final InetAddress getRemoteAddress() {
		return tcpConnection.getRemoteAddress();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final int getRemotePort() {
		return tcpConnection.getRemotePort();
	}
	


	/**
	 * see {@link INonBlockingConnection#setWriteTransferRate(int)} 
	 */
	public final void setWriteTransferRate(int bytesPerSecond) throws ClosedChannelException, IOException {
		tcpConnection.setWriteTransferRate(bytesPerSecond);
	}
	
	
	/**
	 * see {@link INonBlockingConnection#setFlushmode(FlushMode)} 
	 */
	public final void setFlushmode(FlushMode flushmode) {
		tcpConnection.setFlushmode(flushmode);
	}
	
	
	/**
	 * see {@link INonBlockingConnection#getFlushmode(FlushMode)} 
	 */
	final FlushMode getFlushmode() {
		return tcpConnection.getFlushmode();
	}
	
	
	/**
	 * see {@link INonBlockingConnection#getWorkerpool()} 
	 */
	public final Executor getWorkerpool() {
		return tcpConnection.getWorkerpool();
	}
	
	
	
	/**
	 * sets the worker pool 
	 * @param workerpool  the worker pool
	 */
	public void setWorkerpool(Executor workerpool) {
		tcpConnection.setWorkerpool(workerpool);
	}

	
	
	/**
	 * see {@link INonBlockingConnection#activateSecuredMode()} 
	 */
	public void activateSecuredMode() throws IOException {
		tcpConnection.activateSecuredMode();
	}
	
	
	/**
	 * see {@link INonBlockingConnection#isSecure()} 
	 */
	public final boolean isSecure() {
		return tcpConnection.isSecure();
	}
	
	
	/**
	 * checks if the connection is persistent
	 * 
	 * @return true, if the connection is persistent
	 */
	public final boolean isPersistent() {
		return isPersistent.get();
	}
	
	
	/**
	 * set if the connection is persistent 
	 * 
	 * @param persistent true, if the connection is persistent
	 */
	final protected void setPersistent(boolean persistent) {
		isPersistent.set(persistent);
	}

	
	/**
	 * returns if the connection is persistent 
	 * 
	 * @return true, if the connection is persistent
	 */
	final protected boolean getPersistent() {
		return isPersistent.get();
	}

	
	
	/**
	 * {@inheritDoc}
	 */
	public final String getId() {
		try {
			return tcpConnection.getId();
		} catch (Throwable t) {
			return "<unknown>";
		}
	}

	
	/**
	 * {@inheritDoc}
	 */
	public final boolean isOpen() {
		return tcpConnection.isOpen();
	}
	

	/**
	 * {@inheritDoc}
	 */
	public final void close() throws IOException {
		try {
			releaseResources();
			
			if (isReuseable()) {
				tcpConnection.close();

			// destroy (underlying tcp) connection by using connection pool. The underlying connection could be a pooled one)
			// The connection pool detects automatically if the connection is pooled or not. The connection will be 
			// closed (logically) anyway
			} else {
				NonBlockingConnectionPool.destroy(tcpConnection);
			}
		} catch (Exception e) {
		    // eat and log exception
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("error occured by closing htttp connection " + getId() + " " + DataConverter.toString(e));
			}
		}			
	}
	

	/**
	 * closes this connection by swallowing io exceptions
	 */
	protected final void closeSilence() {
		try {
			close();
		} catch (IOException ioe) {
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("[" + getId() + "] error occured by closing connection " + getId() + " " + ioe.toString());
			}
			
			try {
				NonBlockingConnectionPool.destroy(tcpConnection);
			} catch (IOException e) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("[" + getId() + "] error occured by closing connection " + getId() + " " + e.toString());
				}
			}
		}
	}

	
	private void releaseResources() {
		BodyDataSink bds = bodyDataSink.get();
		if (bds != null) {
			bodyDataSink.set(null);
			bds.onUnderlyingHttpConnectionClosed();
		}
	}


	private boolean isReuseable() {
		return isPersistent.get() && !isReceivingSuspended(); 
	}
	

	
	/**
	 * destroy the connection
	 */
	public void destroy() {
		
		isPersistent.set(false);
		
		releaseResources();
		
		// destroy (underlying tcp) connection by using connection pool. The underlying connection could be a pooled one)
		// The connection pool detects automatically if the connection is pooled or not. The connection will be 
		// closed (logically) anyway
		try {
			NonBlockingConnectionPool.destroy(tcpConnection);
		} catch (IOException ioe) {
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("error occured by destroying htttp connection " + getId() + " " + ioe.toString());
			}
		}
	}
	
	

	/**
	 * {@inheritDoc}
	 */
	final public void suspendReceiving() throws IOException {
		if (LOG.isLoggable(Level.FINE)) {
			LOG.fine("suspend receving");
		}
		tcpConnection.suspendReceiving();
		
		countSuspendReceiving.incrementAndGet();
	}


	
	/**
	 * {@inheritDoc}
	 */
	final public boolean isReceivingSuspended() {
		return tcpConnection.isReceivingSuspended();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	final public void resumeReceiving() throws IOException {
		if (LOG.isLoggable(Level.FINE)) {
			LOG.fine("[" + getId() + "] resume receving");
		}

		tcpConnection.resumeReceiving();
		
		countResumeReceiving.incrementAndGet();
	}

	
	final void setBodyDataSink(BodyDataSink bodyChannel) {
		bodyDataSink.set(bodyChannel);
	}
	
	
	final boolean removeBodyDataSink(BodyDataSink bodyChannel) {
		BodyDataSink bds = bodyDataSink.get(); 
		if ((bds != null) && (bds == bodyChannel)) {
			bodyDataSink.set(null);
			return true;
		}
		
		return false;
	}
	
	
	final void flush() throws IOException {
		tcpConnection.flush();
	}
	
	final int write(String txt) throws IOException {
		
		if (LOG.isLoggable(Level.FINE)) {
			LOG.fine("[" + getId() + "] TCP write: " + txt);
		}
		
		lastTimeDataWritten = System.currentTimeMillis();
		
		int size = tcpConnection.write(txt);
		
		lastTimeDataWritten = System.currentTimeMillis();
		countSendBytes += size;
		
		return size;
	}
	
	final long write(ByteBuffer[] buffer, IWriteCompletionHandler completionHandler) throws IOException {
		
		if (LOG.isLoggable(Level.FINE)) {
			ByteBuffer[] bufs = new ByteBuffer[buffer.length];
			for (int i = 0; i < buffer.length; i++) {
				bufs[i] = buffer[i].duplicate();
			}
			
			LOG.fine("[" + getId() + "] TCP write: " + DataConverter.toString(bufs));
		}

		
		
		int size = 0;
		if (completionHandler != null) {
		    for (ByteBuffer byteBuffer : buffer) {
                size += byteBuffer.remaining();
            }
		    tcpConnection.write(buffer, completionHandler);
		    
		} else {
			size = (int) tcpConnection.write(buffer);
		}
		
		lastTimeDataWritten = System.currentTimeMillis();
		countSendBytes += size;
		
		return size;
	}
	
	
	protected abstract IMessageHandler getMessageHandler();
	
	
	protected void onProtocolException(Exception ex) {
		destroy();
	}
	
	
	protected void onConnect() throws IOException {
		
		for (final IHttpConnectionHandler connectionHandler : connectionHandlers) {
			HttpConnectionHandlerInfo connectionHandlerInfo = HttpUtils.getHttpConnectionHandlerInfo(connectionHandler);
			
			if (connectionHandlerInfo.isConnectHandler()) {
				Runnable task = new Runnable() {
					
					public void run() {
						try {
							((IHttpConnectHandler) connectionHandler).onConnect(AbstractHttpConnection.this);
						} catch (IOException ioe) {
							if (LOG.isLoggable(Level.FINE)) {
								LOG.fine("[" + getId() + "] error occured by performing onConnect on " + connectionHandler + " reason: " + ioe.toString());
							}
							destroy();
						}
					}
				};
				
				
				if (connectionHandlerInfo.isConnectHandlerMultithreaded()) {
					multimodeExcutor.processMultithreaded(task);
				} else {
					multimodeExcutor.processNonthreaded(task);
				}
			}
		}
	}

	
	
	protected void onDisconnect() {
		
		try {
			
			if (!isDisconnected.getAndSet(true)) {
				releaseResources();

				for (final IHttpConnectionHandler connectionHandler : connectionHandlers) {
					HttpConnectionHandlerInfo connectionHandlerInfo = HttpUtils.getHttpConnectionHandlerInfo(connectionHandler);
	    				
					if (connectionHandlerInfo.isDisconnectHandler()) {
						callOnDisconnect(connectionHandler, connectionHandlerInfo);
					}
				}
				connectionHandlers.clear();
		    }
			
		} catch (Throwable t) {
			// log and eat exception
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine("[" + getId() + "] error occured by closing http connection " + DataConverter.toString(t));
			}
		}
	}
	
	
	

	
	private void callOnDisconnect(final IHttpConnectionHandler connectionHandler, HttpConnectionHandlerInfo connectionHandlerInfo) {
		Runnable task = new Runnable() {
				
			public void run() {
				try {
					((IHttpDisconnectHandler) connectionHandler).onDisconnect(AbstractHttpConnection.this);
				} catch (IOException ioe) {
					if (LOG.isLoggable(Level.FINE)) {
						LOG.fine("[" + getId() + "] error occured by performing ondisconnect on " + connectionHandler + " reason: " + ioe.toString());
					}
					destroy();
				}
			}
		};
			
		if (connectionHandlerInfo.isDisconnectHandlerMultithreaded()) {
			multimodeExcutor.processMultithreaded(task);
		} else {
			multimodeExcutor.processNonthreaded(task);
		}
	}
	
	
	protected void onConnectionTimeout() throws IOException {
		this.close();
	}
	
	
	protected void onIdleTimeout() throws IOException {
		this.close();
	}

	
	protected final BodyDataSink writeMessage(IHttpHeader header, boolean destroyOnClose) throws IOException {
		lastTimeDataWritten = System.currentTimeMillis();
		
		//  transfer encoding header is set with chunked -> chunked body
		if (isChunkedTransferEncoding(header)) {
			return new BodyDataSink(this, getExecutor(), new FullMessageChunkedWriter(this, destroyOnClose, header), header.getCharacterEncoding());
		}

		
		// contains a non-zero Content-Length header  -> bound body
		int contentLength = header.getContentLength();
		if (contentLength != -1) {
			return writeMessage(header, destroyOnClose, contentLength);
		}
		
		
		// is connection header set with close?  -> SimpleMessage
		if ((header.getConnection() != null) && (header.getConnection().equalsIgnoreCase("close"))) {
			setPersistent(false);
			return new BodyDataSink(this, getExecutor(), new SimpleMessageWriter(this, destroyOnClose, header), header.getCharacterEncoding());
		}
		
		
		// is content-type set?
		if (header.getContentType() != null) {
			setPersistent(false);
			return new BodyDataSink(this, getExecutor(), new SimpleMessageWriter(this, destroyOnClose, header), header.getCharacterEncoding());
		}
	
		
		// return PseudoBody
		return new BodyDataSink(this, getExecutor(), new BodylessMessageWriter(this, destroyOnClose, header), header.getCharacterEncoding());
	}
	
	
	protected final BodyDataSink writeMessage(IHttpHeader header, boolean destroyOnClose, int length) throws IOException {
		lastTimeDataWritten = System.currentTimeMillis();
		
		// contains a non-zero Content-Length header  -> bound body 
		if (length > 0) {
			return new BodyDataSink(this, getExecutor(), new FullMessageWriter(this, destroyOnClose, header, length), header.getCharacterEncoding());
			
		} else {
			header.removeHeader("Content-Type");
			// return PseudoBody
			return new BodyDataSink(this, getExecutor(), new BodylessMessageWriter(this, destroyOnClose, header), header.getCharacterEncoding());
		}
	}
	
	
	protected final void writeMessage(IHttpMessage message, boolean destroyOnClose) throws IOException {
		BodyDataSink bodyDataSink = writeMessage(message.getMessageHeader(), destroyOnClose);
		NonBlockingBodyDataSource bodyDataSource = message.getNonBlockingBody();
		
		// bodyless message?
		if (message.getNonBlockingBody() == null) {
			FlushMode flushMode = getFlushmode();
			setFlushmode(FlushMode.ASYNC);
			write(message.getMessageHeader().toString() + "\r\n");
			flush();
			
			setFlushmode(flushMode);
			return;
			
		// .. no
		} else {
			bodyDataSink.setFlushmode(FlushMode.ASYNC);
			bodyDataSink.setAutoflush(false);
			
			// is body complete?
			if (bodyDataSource.isComplete()) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("[" + getId() + "] message body to sent is complete.writing all data to body data sink ");
				}
						
				int available = bodyDataSource.available();
				if (available > 0) {
					bodyDataSink.write(bodyDataSource.readByteBufferByLength(available));
				}
				bodyDataSink.close();	
			
			// .. no
			} else {
				BodyForwarder forwarder = new BodyForwarder(bodyDataSource, bodyDataSink);
				bodyDataSource.setDataHandler(forwarder);
				bodyDataSink.flush();
			}
		}
	}
	
	
	

	static boolean isChunkedTransferEncoding(IHttpHeader header) {
		
		String transferEncoding = header.getTransferEncoding();
		if ((transferEncoding != null) && (transferEncoding.equalsIgnoreCase("chunked"))) {
			return true;
		}
		
		return false;
	}

	
	
	/**
	 * returns the request handler info
	 * @param requestHandler the request handler 
	 * @return the request handler info
	 */
	protected static RequestHandlerInfo getRequestHandlerInfo(IHttpRequestHandler requestHandler) {
		return HttpUtils.getHttpRequestHandlerInfo(requestHandler);
	}
	
	
	/**
	 * set the http connection 
	 * 
	 * @param responseHandler  the future response handler
	 * @param httpConnection   the http connection
	 */
	protected static void setUnderlyingResource(FutureResponseHandler responseHandler, IDestroyable detroyable) {
	    responseHandler.setUnderlyingResource(detroyable);
	}

		
	/**
	 * returns the response handler info
	 * @param responseHandler the response handler 
	 * @return the response handler info
	 */
	protected static ResponseHandlerInfo getResponseHandlerInfo(IHttpResponseHandler responseHandler) {
		return HttpUtils.getHttpResponseHandlerInfo(responseHandler);
	}

	/**
	 * returns true if the requests will accept a chunked body 
	 * @param request the request 
	 * @return true if the requests will accept a chunked body
	 */
	protected static boolean isAcceptingChunkedResponseBody(IHttpRequest request) {
		
		String protocolVersion = request.getProtocolVersion();
		
		if (protocolVersion.equals("1.1")) {
			return true;
			
		} else {
			int idx = protocolVersion.indexOf(".");
			
			int minor = Integer.parseInt(protocolVersion.substring(idx + 1, protocolVersion.length()));
			if (minor > 0) {
				return true;
			}
			
			int major = Integer.parseInt(protocolVersion.substring(0, idx));
			if (major > 1) {
				return true;
			}
		}
		return false;
	}
	
	
	/**
	 * schedule a time task 
	 * 
	 * @param task    the timer task
	 * @param delay   the delay 
	 * @param period   the period
	 */
	protected static void schedule(TimerTask task, long delay, long period) {
		TIMER.schedule(task, delay, period);
	}
	
	

	/**
	 * creates a new DataSuorceSinkPair
	 * 
	 * @param connection    the http connection
	 * @param executor      a executor
	 * @param encoding      the body encoding
	 * @return a new DataSuorceSinkPair
	 * @throws IOException if an exception occurs
	 */
	protected static final DataSourceSinkPair newBodyDataSourceSinkPair(AbstractHttpConnection connection, IMultimodeExecutor executor, String encoding) throws IOException {
		final NonBlockingBodyDataSource bodyDataSource = new NonBlockingBodyDataSource(BodyType.IN_MEMORY, encoding);
		
		IMessageWriter messageWriter = new IMessageWriter() {
				
		    public void flush(ByteBuffer[] bodyData, boolean isContentImmutable, FlushMode flushMode, IWriteCompletionHandler completionHandler) throws IOException {
		    	if (!bodyDataSource.isOpen()) {
		    		throw new ClosedChannelException();
		    	}
				bodyDataSource.append(isContentImmutable, bodyData, completionHandler);
			}
				
			public void close() throws IOException {
				bodyDataSource.setComplete(true);
			}
				
			public void destroy() {
				bodyDataSource.destroy();
			}
			
			public boolean isNetworkEndpoint() {
				return false;
			}
						
			public int getPendingWriteDataSize() {
				return 0;
			}
		};
		
		
		final BodyDataSink bodyDataSink = new BodyDataSink(connection, executor, messageWriter, encoding);
		
		IBodyCloseListener closeListener = new IBodyCloseListener() {
			
			@Execution(Execution.NONTHREADED)
			public void onClose() throws IOException {
				bodyDataSink.close();
			}
		};
		
		bodyDataSource.addCloseListener(closeListener);


		return new DataSourceSinkPair(bodyDataSource, bodyDataSink);
	}
	

	/**
	 * wraps the request by a request object which will read parameters from the body 
	 * @param request the request to wrap 
	 * @return request object which will read parameters from the body
	 * @throws IOException if an exception occurs 
	 */
	protected static IHttpRequest newFormEncodedRequestWrapper(IHttpRequest request) throws IOException {
		return HttpUtils.newFormEncodedRequestWrapper(request);
	}
	 
	
		
	/**
	 * register the task as which will be executed if the body data sink will be closed 
	 * @param bodyDataSink the body data sink
	 * @param task the task to execute
	 */
	protected final void setBodyCloseListener(BodyDataSink bodyDataSink, Runnable task) {
		bodyDataSink.addCloseListener(new BodyCloseListener(task));
	}
	
	
	@Execution(Execution.NONTHREADED)
	private static final class BodyCloseListener implements IBodyCloseListener {
		
		private Runnable task = null;
		
		public BodyCloseListener(Runnable task) {
			this.task = task;
		}
		
		public void onClose() throws IOException {
			task.run();
		}
	}
	
	
	
	/**
	 * initiates that the body data source will be destroyed, if the message body is received   
	 * 
	 * @param bodyDataSource                     the body data source
	 * @param isCloseConnectionAfterReceived     true, if the body data source will be destroyed 
	 */
	protected final void setDestroyConnectionAfterReceived(NonBlockingBodyDataSource bodyDataSource, boolean isCloseConnectionAfterReceived) {
		bodyDataSource.setDestroyAfterReceived(isCloseConnectionAfterReceived);
	}
	
	
	/**
	 * initiates that the body data source will be closed, if the message body is received   
	 * 
	 * @param bodyDataSource                     the body data source
	 * @param isCloseConnectionAfterReceived     true, if the body data source will be closed
	 */
	protected final void setCloseConnectionAfterReceived(NonBlockingBodyDataSource bodyDataSource, boolean isCloseConnectionAfterReceived) {
		bodyDataSource.setCloseAfterReceived(isCloseConnectionAfterReceived);
	}
	
	

	/**
	 * creates a new body data sink
	 * 
	 * @param bodyDataHandler  the body handler 
	 * @param encoding         the body encoding
	 * @return a new body data sink 
	 * @throws IOException in an exception occurs
	 */
	protected final BodyDataSink newBufferedBodyDataSink(IBodyDataHandler bodyDataHandler, String encoding) throws IOException {
		
		return new BodyDataSink(this, getExecutor(), new MessageWriter(bodyDataHandler, encoding), encoding);
	}

	
	private static final class MessageWriter implements IMessageWriter {
		
		private NonBlockingBodyDataSource bodyDataSource = null;
		private IBodyDataHandler bodyDataHandler = null;
		
		public MessageWriter(IBodyDataHandler bodyDataHandler, String encoding) {
			this.bodyDataHandler = bodyDataHandler;
			bodyDataSource = new NonBlockingBodyDataSource(BodyType.IN_MEMORY, encoding);
		}
		
		public void flush(ByteBuffer[] bodyData, boolean isContentImmutable, FlushMode flushMode, IWriteCompletionHandler completionHandler) throws IOException {
			bodyDataSource.append(isContentImmutable, bodyData, completionHandler);
			bodyDataHandler.onData(bodyDataSource);
		}
		
		public void close() throws IOException {
			bodyDataSource.close();
			bodyDataHandler.onData(bodyDataSource);
		}
		
		public void destroy() {
			bodyDataSource.destroy();
			bodyDataHandler.onData(bodyDataSource);
		}
		
		public boolean isNetworkEndpoint() {
			return false;
		}
		
		public int getPendingWriteDataSize() {
			return 0;
		}
	}; 
	

	/**
	 * return the max write buffer size
	 * 
	 * @return the max write buffer size
	 */
	protected static int getMaxWriteBufferSize() {
		return MAX_WRITEBUFFER_SIZE;
	}
	
	
	/**
	 * generates a system error html
	 * 
	 * @param errorCode      the error code 
	 * @param msg            the error message
	 * @param connectionId   the connection id
	 * @return the html page
	 */
	protected static String generateErrorMessageHtml(int errorCode, String msg, String connectionId) {
		
		if (msg == null) {
			msg = HttpUtils.getReason(errorCode);
		}
		
		
		msg = msg.replace("\r", "<br/>");

		String txt = "<html>\r\n" +
		             "  <!-- This page is auto-generated by xLightweb (http://xLightweb.org) -->\r\n" +
		             "  <!-- id " + connectionId + " -->\r\n" +
		             "  <!-- xLightweb/"  + HttpUtils.getImplementationVersion() +  " (xSocket/" + ConnectionUtils.getImplementationVersion() + ") -->\r\n" +
					 "  <head>\r\n" +
					 "    <title>Error " + errorCode + "</title>\r\n" +
					 "    <meta http-equiv=\"cache-control\" content=\"no-cache\"/>" +
					 "  </head>\r\n\r\n" +
					 "  <body>\r\n" +
					 "    <H1 style=\"color:#0a328c;font-size:1.5em;\">ERROR " + errorCode + "</H1>\r\n" +
					 "    <p style=\"font-size:1.5em;\">" + msg + "</p>\r\n" +
					 "    <p style=\"font-size:0.8em;\">" + new SimpleDateFormat("EEE, d MMM yyyy HH:mm:ss Z").format(new Date()) + "    xLightweb (" + HttpUtils.getImplementationVersion() + ")</p>\r\n" +
					 "  <body>\r\n" +
					 "</html>\r\n";
			
		return txt;
	}
	
	
	
	
	/**
	 * creates a empty body data sink
	 * @return the body data sink
	 * @throws IOException in an exception occurs
	 */
	protected final BodyDataSink newEmtpyBodyDataSink() throws IOException {
		
		return new BodyDataSink(this, getExecutor(), NULL_BODYWRITER, "UTF-8");
	}

	
	private static final class NullMessageWriter implements IMessageWriter {
		
	    public void flush(ByteBuffer[] bodyData, boolean isContentImmutable, FlushMode flushMode, IWriteCompletionHandler completionHandler) throws IOException {
		}
		
		public void close() throws IOException {
		}
		
		public void destroy() {
		}
		
		public boolean isNetworkEndpoint() {
			return false;
		}
		
		public int getPendingWriteDataSize() {
			return 0;
		}
	}

	
	
	
	private AbstractHttpProtocolHandler getProtocolHandler() {
		return protocolHandler;
	}
	
	
	private static AbstractHttpConnection getHttpConnection(INonBlockingConnection tcpConnection) {
		return (AbstractHttpConnection) tcpConnection.getAttachment();
	}
	
	
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append(getId() + " " + tcpConnection.getLocalAddress() + ":" + tcpConnection.getLocalPort() + " -> " + tcpConnection.getRemoteAddress() + ":" + tcpConnection.getRemotePort());
		if (!tcpConnection.isOpen()) {
			sb.append("  (closed)");
		}
		
		return sb.toString();
	}
	
	
	/**
	 * message handler definition 
	 * 
	 * @author grro
	 */
	protected interface IMessageHandler {

		/**
		 * call back if a message is received
		 * 
		 * @param message       the message
		 * @throws IOException if an exception occurs 
		 */
		public void onMessage(IHttpMessage message) throws IOException;
		
		
		/**
		 * call back if an exception is occurred 
		 * @param ioe the exception
		 */
		public void onException(IOException ioe);
		
		
		/**
		 * returns true if a body less message is expected  
		 * @return true if a body less message is expected 
		 */
		public boolean isBodylessMessageExpected();
	}
	

	/**
	 * multimode executor definition 
	 * @author grro
	 */
	protected interface IMultimodeExecutor {

		/**
		 * process the task multi threaded
		 * @param task the task to process
		 */
		void processMultithreaded(Runnable task);
		
		
		/**
		 * process the task non threaded
		 * @param task the task to process
		 */
		void processNonthreaded(Runnable task);
	}
	
	
	/**
	 * data source sink pair 
	 * @author grro
	 */
	protected static final class DataSourceSinkPair {
		
		private NonBlockingBodyDataSource bodyDataSource = null;
		private BodyDataSink bodyDataSink = null;
		
		DataSourceSinkPair(NonBlockingBodyDataSource bodyDataSource, BodyDataSink bodyDataSink) {
			this.bodyDataSource = bodyDataSource;
			this.bodyDataSink = bodyDataSink;
		}
		
		/**
		 * returns the body data source
		 * @return the body data source
		 */
		public NonBlockingBodyDataSource getBodyDataSource() {
			return bodyDataSource;
		}

		/**
		 * returns the body data sink
		 * @return the body data sink
		 */
		public BodyDataSink getBodyDataSink() {
			return bodyDataSink;
		}	
	}

	
	
	private final class MultimodeExecutor implements IMultimodeExecutor {
		
		private final SerializedTaskQueue taskQueue = new SerializedTaskQueue();

		
		public void processMultithreaded(Runnable task) {
			taskQueue.performMultiThreaded(task, getWorkerpool());
		}
		
		public void processNonthreaded(Runnable task) {
			taskQueue.performNonThreaded(task, getWorkerpool());
		}
		
		
		
	}
	
	
	@Execution(Execution.NONTHREADED)
	private final class DataHandler implements IDataHandler, IConnectHandler, IDisconnectHandler, IIdleTimeoutHandler, IConnectionTimeoutHandler {
		

		private final ComposedByteBuffer rawData = new ComposedByteBuffer();

		
		public boolean onData(INonBlockingConnection connection) throws BufferUnderflowException {
			lastTimeDataReceivedMillis = System.currentTimeMillis();
			
			if (connection.isOpen()) {
				AbstractHttpProtocolHandler protocolHandler = (AbstractHttpProtocolHandler) getHttpConnection(connection).getProtocolHandler();
					
				try {
					
					// copying available network data into raw data buffer
					int available = connection.available();
					if (available > 0) {
						ByteBuffer[] data = connection.readByteBufferByLength(available);
		
						if (LOG.isLoggable(Level.FINE)) {
							ByteBuffer[] bufs = new ByteBuffer[data.length];
							for (int i = 0; i < data.length; i++) {
								bufs[i] = data[i].duplicate();
							}
							LOG.fine("[" + getId() + "] TCP read: " + DataConverter.toString(bufs));
						}
							
						rawData.append(data);
								
							
						// handle data based on the buffer
						do {
							protocolHandler.onData(AbstractHttpConnection.this, connection, rawData);
						} while (!rawData.isEmpty());
						
						rawData.clear();
					}
					
					return true;
						
				} catch (ClosedChannelException cce) {
					setPersistent(false);
					protocolHandler.onDisconnect();
						
				} catch (IOException ex) {
				    setPersistent(false);
                    protocolHandler.onException(AbstractHttpConnection.this, ex, rawData);
					onProtocolException(ex);
				}
			} 
			
			return true;
		}
		
		
		public boolean onConnect(INonBlockingConnection connection) throws IOException, BufferUnderflowException, MaxReadSizeExceededException {
			return true;
		}
		
		
		public boolean onDisconnect(INonBlockingConnection connection) throws IOException {
            if (LOG.isLoggable(Level.FINE)) {
                LOG.fine("[" + getId() + "] disconnected");
            }
            	
            AbstractHttpProtocolHandler protocolHandler = (AbstractHttpProtocolHandler) getHttpConnection(connection).getProtocolHandler();
            protocolHandler.onDisconnect();
			
			AbstractHttpConnection.this.onDisconnect();
			return true;
		}
		
		
		public boolean onConnectionTimeout(INonBlockingConnection connection) throws IOException {
			AbstractHttpConnection.this.onConnectionTimeout();
			return true;
		}
		
		
		public boolean onIdleTimeout(INonBlockingConnection connection) throws IOException {
			AbstractHttpConnection.this.onIdleTimeout();
			return true;
		}
	}		
}
