/*
 *  Copyright (c) xlightweb.org, 2008 - 2009. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;


import java.io.IOException;




/**
 * Header line parser
 * 
 * @author grro@xlightweb.org
 */
final class HeaderlineParser {
	

	private static final char[] ISO_8859_1_ARRAY = HttpUtils.getISO_8859_1_Array();
	
	private static final byte CR = 13;
	private static final byte LF = 10;
	private static final byte SPACE = 32;
	private static final byte HTAB = 9;
	
	
		
	private static ThreadLocal<HeaderlineParser> instanceThreadLocal = new ThreadLocal<HeaderlineParser>();

	private int state = 0;
	
	private StringBuilder nameBuilder = new StringBuilder(26);
	private StringBuilder valueBuilder = new StringBuilder();
	private int posLastCharNotSpaceOrHtab; 
	private boolean isPreviousCharLF = true;
	

	
	/**
	 * returns a new instance 
	 * 
	 * @return new instance 
	 */
	static HeaderlineParser newInstance() {
		HeaderlineParser instance = instanceThreadLocal.get();
		if (instance == null) {
			instance = new HeaderlineParser();
		}
		
		instanceThreadLocal.set(null);
		return instance;
	}
	 

	
	/**
	 * recycle an instance
	 * 
	 * @param parser the instance to recycle  
	 */
	static void recycleInstance(HeaderlineParser parser) {
		instanceThreadLocal.set(parser);
	}
	
	
	private HeaderlineParser() { 	}
	

	
	/**
	 * parses the header 
	 * 
	 * @param composedByteBuffer    the raw network data input buffer 
	 * @param messageHeader         the message header
	 * @return true if the header is read
	 * @throws BadMessageException if the message is invalid
	 * @throws IOException if an exception occurs
	 */
	public boolean parse(ComposedByteBuffer composedByteBuffer, AbstractHttpHeader messageHeader) throws BadMessageException, IOException {
		return parse(composedByteBuffer, (IHeaderSink) messageHeader);
	}
	

	/**
	 * parses the header 
	 * 
	 * @param composedByteBuffer   the raw network data input buffer
	 * @param headerSink           the data sink
	 * @return true if the header is read
	 * @throws IOException if an exception occurs
	 */
	boolean parse(ComposedByteBuffer composedByteBuffer, IHeaderSink headerSink) throws IOException {
		
		byte b = composedByteBuffer.getByte();
		
		
		
		// RFC 2616:
		// CHAR           = <any US-ASCII character (octets 0 - 127)>
		// token          = 1*<any CHAR except CTLs or separators>
		// The TEXT rule is only used for descriptive field contents and values
		// that are not intended to be interpreted by the message parser. Words
		// of *TEXT MAY contain characters from character sets other than ISO-
		// 8859-1 [22] only when encoded according to the rules of RFC 2047
		// 
		// message-header = field-name ":" [ field-value ]
		// field-name     = token
		// field-value    = *( field-content | LWS )
		// field-content  = <the OCTETs making up the field-value
		//                   and consisting of either *TEXT or combinations
		//                   of token, separators, and quoted-string>
		
		
		// filter CR:
		// RFC 2616 (19.3 Tolerant Applications)
		// ... The line terminator for message-header fields is the sequence CRLF.
		// However, we recommend that applications, when parsing such headers,
		// recognize a single LF as a line terminator and ignore the leading CR...


		switch (state) {
		
		    // reading first char of line
			case 0:
				switch (b) {
					case CR:
						break;
					
					case LF:
						// header end? 
						if (isPreviousCharLF) {
							addLine(headerSink);   
							reset();
							return true;
						} 
						isPreviousCharLF = true;
						break;
						
				
					case SPACE:
						valueBuilder.append(" ");
						state = 10; 
						break;
						
					case HTAB:
						valueBuilder.append(" ");
						state = 10;
						break;
						
						
					default:
						isPreviousCharLF = false;
						addLine(headerSink);  // add previous read line (if exists) 
					
						nameBuilder.append((char) b);
						state = 5;
						break;
				}
				break;

				
				
		    // reading remaining char of name
			case 5:
				switch (b) {
					case CR:
						break;
					
					case LF:
						state = 0;
						isPreviousCharLF = true;
						break;
						
				
					case SPACE:
						state = 10; 
						break;
						
					case HTAB:
						state = 10;
						break;

						
					case ':':
						state = 10;
						break;
						
						
					default:
						nameBuilder.append((char) b);
						break;
				}
				break;

				
				
			// waiting until value starts 
			case 10:
				switch (b) {
					case SPACE:
						break;
						
					case HTAB:
						break;
						
					default:
						valueBuilder.append(toISO_8859_1_Char(b));
					    posLastCharNotSpaceOrHtab = valueBuilder.length();
						state = 11;
						break;
				}
				
				break;

				
			
			case 11:
				switch (b) {
					case CR:
						break;
					
					case LF:
						isPreviousCharLF = true;
						state = 0;
						break;
						
					case SPACE:
						valueBuilder.append(toISO_8859_1_Char(b)); 
						break;
						
					case HTAB:
						valueBuilder.append(toISO_8859_1_Char(b));
						break;
						
					default:
						valueBuilder.append(toISO_8859_1_Char(b));
						posLastCharNotSpaceOrHtab = valueBuilder.length();
						break;
				}
			
				break;
		}
		
		return false;
	}
	
	
	/**
	 * reset
	 */
	public void reset() {
		state = 0;
		
		nameBuilder.setLength(0);
		valueBuilder.setLength(0);
		posLastCharNotSpaceOrHtab = 0;
		
		isPreviousCharLF = true;
	}
	
	
	
	/**
	 * recycles the parser
	 */
	public void recycle() {
		HeaderlineParser.recycleInstance(this);
	}
	
	
	
		
	private void addLine(IHeaderSink headerSink) {
		if ((nameBuilder.length() > 0) && (valueBuilder.length() > 0)) {
			valueBuilder.setLength(posLastCharNotSpaceOrHtab);  // truncate trailing spaces
			headerSink.addHeader(nameBuilder.toString(), valueBuilder.toString());
			nameBuilder.setLength(0);
			valueBuilder.setLength(0);
		}
	}

	
	
	char toISO_8859_1_Char(byte b) {
		int i = b;
		i &= 0x000000FF;
		
		return ISO_8859_1_ARRAY[i]; 
	}
	
	
	
	static interface IHeaderSink {
		
		public void addHeader(String headername, String headervalue);
	}
}
