/*
 *  Copyright (c) xlightweb.org, 2008 - 2009. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.net.URLEncoder;
import java.nio.ByteBuffer;
import java.util.Enumeration;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;



/**
 * http request 
 *
 * @author grro@xlightweb.org
 */
public class HttpRequest extends AbstractHttpMessage implements IHttpRequest {
	
	private static final Logger LOG = Logger.getLogger(HttpRequest.class.getName());
	
	
	private final IHttpRequestHeader requestHeader;
	
	private String requestHandlerPath = "";
	private String contextPath = ""; 
	

	
	
	/**
	 * constructor
	 * 
	 * @param requestHeader  the request header
	 */
	public HttpRequest(IHttpRequestHeader requestHeader) {
		this.requestHeader = requestHeader;
		
		if (!isBodylessRequestMethod(requestHeader.getMethod())) {
			requestHeader.setContentLength(0);
		}
	}
	

	/**
	 * constructor
	 * 
	 * @param requestHeader    the request header 
	 * @param bodyDataSource   the body data source
	 * 
	 * @throws IOException if an exception occurs  
	 */
	public HttpRequest(IHttpRequestHeader requestHeader, NonBlockingBodyDataSource bodyDataSource) throws IOException {
		this.requestHeader = requestHeader;
		setBodyDataSource(bodyDataSource);
	}
	
	

	/**
	 * constructor
	 * 
	 * @param requestHeader    the request header 
	 * @param bodyData         the body data 
	 * 
	 * @throws IOException if an exception occurs  
	 */
	public HttpRequest(IHttpRequestHeader requestHeader, ByteBuffer[] bodyData) throws IOException {
		this.requestHeader = requestHeader;
		setBodyDataSource(bodyData, requestHeader.getCharacterEncoding());
	}
	
	
	/**
 	 * constructor
	 * 
	 * @param method        the method (GET, POST, ...)
	 * @param url           the url
	 * @param contentType   the content type
	 * @param body          the body 
	 * 
	 * @throws IOException if an exception occurs
	 */
	public HttpRequest(String method, String url, String contentType, String body) throws IOException {
		requestHeader = new HttpRequestHeader(method, url, contentType);
		
		try {
			int length = setBodyDataSource(body, requestHeader.getCharacterEncoding());
			requestHeader.setContentLength(length);
		} catch (UnsupportedEncodingException use) {
			throw new RuntimeException(use.toString());
		}
	}


	
	/**
 	 * constructor
	 * 
	 * @param method        the method (GET, POST, ...)
	 * @param url           the url
	 * @param contentType   the content type
	 * @param encoding      the encoding
 	 * @param body          the body 
	 * 
	 * @throws IOException if an exception occurs
	 */
	public HttpRequest(String method, String url, String contentType, String encoding,  String body) throws IOException {

		if (HttpUtils.parseEncoding(contentType) == null) {
			contentType = contentType + "; charset=" + encoding;
		}
		
		requestHeader = new HttpRequestHeader(method, url, contentType);
		
		try {
			int length = setBodyDataSource(body, encoding);
			requestHeader.setContentLength(length);
		} catch (UnsupportedEncodingException use) {
			throw new RuntimeException(use.toString());
		}
	}
	
	
	/**
 	 * constructor
	 * 
	 * @param method        the method (GET, POST, ...)
	 * @param url           the url
	 * @param contentType   the content type
	 * @param body          the body 
	 *
	 * @throws IOException if an exception occurs
	 */
	public HttpRequest(String method, String url, String contentType, byte[] body) throws IOException {
		requestHeader = new HttpRequestHeader(method, url, contentType);
		
		try {
			setBodyDataSource(new ByteBuffer[] { ByteBuffer.wrap(body) }, requestHeader.getCharacterEncoding());
		} catch (UnsupportedEncodingException use) {
			throw new RuntimeException(use.toString());
		}
	}
	
	/**
 	 * constructor
	 * 
	 * @param method        the method (GET, POST, ...)
	 * @param url           the url
	 * @param contentType   the content type
	 * @param body          the body 
	 * 
	 * @throws IOException if an exception occurs
	 */
	public HttpRequest(String method, String url, String contentType, ByteBuffer[] body) throws IOException {
		requestHeader = new HttpRequestHeader(method, url, contentType);
		
		try {
			setBodyDataSource(body, requestHeader.getCharacterEncoding());
		} catch (UnsupportedEncodingException use) {
			throw new RuntimeException(use.toString());
		}
	}
	
	
	
	/**
 	 * constructor
	 * 
	 * @param method        the method (GET, POST, ...)
	 * @param url           the url
	 * 
	 */
	public HttpRequest(String method, String url) {
		requestHeader = new HttpRequestHeader(method, url);	
		
		if (!isBodylessRequestMethod(requestHeader.getMethod())) {
			requestHeader.setContentLength(0);
		}
	}
	
	
	
	

	/**
	 * returns the request header
	 *  
	 * @return the request header
	 */
	public IHttpRequestHeader getRequestHeader() {
		return requestHeader;
	}
	
	

	/**
	 * {@inheritDoc}
	 */
	public void addHeaderLine(String line) {
		getRequestHeader().addHeaderLine(line);
	}
		
	
	/**
	 * {@inheritDoc}
	 */
	public String getRequestHandlerPath() {
		return requestHandlerPath;
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public void setRequestHandlerPath(String requestHandlerPath) {
		this.requestHandlerPath = requestHandlerPath;
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public String getContextPath() {
		return contextPath;
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public void setContextPath(String contextPath) {
		this.contextPath = contextPath;
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public void setMethod(String method) {
		getRequestHeader().setMethod(method);
	}
	

	/**
	 * {@inheritDoc}
	 */
	public String getMethod() {
		return getRequestHeader().getMethod();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public List<ContentType> getAccept() {
		return getRequestHeader().getAccept();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public void setHost(String host) {
		getRequestHeader().setHost(host);
	}
	

	/**
	 * {@inheritDoc}
	 */
	public String getServerName() {
		return getRequestHeader().getServerName();
	}
	

	/**
	 * {@inheritDoc}
	 */
	public int getServerPort() {
		return getRequestHeader().getServerPort();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public String getProtocol() {
		return getRequestHeader().getProtocol();
	}
	

	/**
	 * {@inheritDoc}
	 */
	public String getProtocolVersion() {
		return getRequestHeader().getProtocolVersion();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public URL getRequestUrl() {
		return getRequestHeader().getRequestUrl();
	}
	
	

	/**
	 * {@inheritDoc}
	 */
	public void setRequestUrl(URL url) {
		getRequestHeader().setRequestUrl(url);
	}
	
	

	/**
	 * {@inheritDoc}
	 */
	public String getRemoteHost() {
		return getRequestHeader().getRemoteHost();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public String getRemoteAddr() {
		return getRequestHeader().getRemoteAddr();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public int getRemotePort() {
		return getRequestHeader().getRemotePort();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public String getRequestURI() {
		return getRequestHeader().getRequestURI();
	}

	
	public void setRequestURI(String requestUri) {
		getRequestHeader().setRequestURI(requestUri);
	}

	
	/**
	 * {@inheritDoc}
	 */
	public String getPathInfo() {
		return getPathInfo(false);
	}
 
		
	/**
	 * {@inheritDoc}
	 */
	public String getPathInfo(boolean removeSurroundingSlashs) {
		String uri = getRequestURI();
		uri = uri.substring(getContextPath().length(), uri.length());
		uri = uri.substring(getRequestHandlerPath().length(), uri.length()).trim();
		
		if (removeSurroundingSlashs) {
			if (uri.startsWith("/")) {
				uri = uri.substring(1,uri.length());
			}
			
			if (uri.endsWith("/")) {
				uri = uri.substring(0, uri.length() - 1).trim();
			}
			
		} else {
			if (!uri.startsWith("/")) {
				uri = "/"  + uri.substring(1,uri.length());
			}
		}

		return uri;
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public String getConnection() {
		return getRequestHeader().getConnection();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public void setConnection(String connection) {
		getRequestHeader().setConnection(connection);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public String getKeepAlive() {
		return getRequestHeader().getKeepAlive();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public void setKeepAlive(String keepAlive) {
		getRequestHeader().setKeepAlive(keepAlive);		
	}
	
	
	
	/**
	 * {@inheritDoc}
	 */
	public String getHost() {
		return getRequestHeader().getHost();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public String getUpgrade() {
		return getRequestHeader().getUpgrade();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public void setUpgrade(String upgrade) {
		getRequestHeader().setUpgrade(upgrade);		
	}
	
	/**
	 * {@inheritDoc}
	 */
	public String getUserAgent() {
		return getRequestHeader().getUserAgent();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public void setUserAgent(String userAgent) {
		getRequestHeader().setUserAgent(userAgent);
		
	}
	
	/**
	 * {@inheritDoc}
	 */
	public String getQueryString() {
		return getRequestHeader().getQueryString();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public boolean isSecure() {
		return getRequestHeader().isSecure();
	}
	

	
	/**
	 * {@inheritDoc}
	 */
	@SuppressWarnings("unchecked")
	public Enumeration getParameterNames() {
		return getRequestHeader().getParameterNames();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public Set<String> getParameterNameSet() {
		return getRequestHeader().getParameterNameSet();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public String getParameter(String name) {
		return getRequestHeader().getParameter(name);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public String[] getParameterValues(String name) {
		return getRequestHeader().getParameterValues(name);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public String getRequiredStringParameter(String name) throws BadMessageException {
		return getRequestHeader().getRequiredStringParameter(name);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public Integer getIntParameter(String name) throws BadMessageException{
		return getRequestHeader().getIntParameter(name);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public int getRequiredIntParameter(String name) throws BadMessageException{
		return getRequestHeader().getRequiredIntParameter(name);
	}


	/**
	 * {@inheritDoc}
	 */
	public int getIntParameter(String name, int defaultVal) {
		return getRequestHeader().getIntParameter(name, defaultVal);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public Long getLongParameter(String name) throws BadMessageException {
		return getRequestHeader().getLongParameter(name);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public long getRequiredLongParameter(String name) throws BadMessageException {
		return getRequestHeader().getRequiredLongParameter(name);
	}

	
	
	/**
	 * {@inheritDoc}
	 */
	public long getLongParameter(String name, long defaultVal) {
		return getRequestHeader().getLongParameter(name, defaultVal);
	}

	
	/**
	 * {@inheritDoc}
	 */
	public Double getDoubleParameter(String name) throws BadMessageException {
		return getRequestHeader().getDoubleParameter(name);
	}

	
	/**
	 * {@inheritDoc}
	 */
	public double getRequiredDoubleParameter(String name) throws BadMessageException {
		return getRequestHeader().getRequiredDoubleParameter(name);
	}

	
	
	/**
	 * {@inheritDoc}
	 */
	public double getDoubleParameter(String name, double defaultVal) {
		return getRequestHeader().getDoubleParameter(name, defaultVal);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public Float getFloatParameter(String name) throws BadMessageException {
		return getRequestHeader().getFloatParameter(name);
	}
	

	/**
	 * {@inheritDoc}
	 */
	public float getRequiredFloatParameter(String name) throws BadMessageException {
		return getRequestHeader().getRequiredFloatParameter(name);
	}

	
	
	/**
	 * {@inheritDoc}
	 */
	public float getFloatParameter(String name, float defaultVal) {
		return getRequestHeader().getFloatParameter(name, defaultVal);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public Boolean getBooleanParameter(String name) {
		return getRequestHeader().getBooleanParameter(name);
	}

	
	/**
	 * {@inheritDoc}
	 */
	public boolean getRequiredBooleanParameter(String name) throws BadMessageException {
		return getRequestHeader().getRequiredBooleanParameter(name);
	}


	/**
	 * {@inheritDoc}
	 */
	public boolean getBooleanParameter(String name, boolean defaultVal) {
		return getRequestHeader().getBooleanParameter(name, defaultVal);
	}
	
	
	
	/**
	 * {@inheritDoc}
	 */
	public void setParameter(String parameterName, String parameterValue) {
		getRequestHeader().setParameter(parameterName, parameterValue);
	}
	

	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public IHttpHeader getMessageHeader() {
		return requestHeader;
	}	
	
	
	/**
	 * creates a form url encoded body
	 * 
	 * @param formParameters the parameters
	 * @return the form url encoded body
	 */
	static String createFormUrlEncodedBody(String[] formParameters, String encoding) {
		StringBuilder sb = new StringBuilder();
		
		for (String param : formParameters) {
			try {
			    int idx = param.indexOf("=");
				sb.append(URLEncoder.encode(param.substring(0, idx).trim(), encoding) + "=" + URLEncoder.encode(param.substring(idx + 1, param.length()).trim(), encoding) + "&");
			} catch (UnsupportedEncodingException usec) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("error occured by encoding param " + param + " " + usec.toString());
				}
			}
		}
		
		String body = sb.toString();
		if (body.length() > 0) {
			body = body.substring(0, body.length() - 1);
		}
		
		return body;
	}
	

	/**
	 * return true if the method indicated a body less message 
	 * 
	 * @param method the method name
	 * @return  true if the method indicated a body less message
	 */
	static boolean isBodylessRequestMethod(String method) {
		
		if (method.equals(IHttpMessage.GET_METHOD)) {
			return true;
		}
		
		if (method.equals(IHttpMessage.POST_METHOD)) {
			return false;
		}
		
		
		if (method.equals(IHttpMessage.CONNECT_METHOD) || 
			method.equals(IHttpMessage.HEAD_METHOD) || 
		    method.equals(IHttpMessage.TRACE_METHOD)|| 
		    method.equals(IHttpMessage.DELETE_METHOD)|| 
		    method.equals(IHttpMessage.OPTIONS_METHOD)) {
			
			return true;
		}
		
		return false;
	}
}
