/*
 *  Copyright (c) xlightweb.org, 2008 - 2009. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;


import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xsocket.connection.INonBlockingConnection;

import static org.xlightweb.HttpUtils.*;

/**
 * http response header parser
 *   
 * @author grro@xlightweb.org
 */
final class HttpResponseHeaderParser implements IMessageHeaderParser {
	
    
    private static final Logger LOG = Logger.getLogger(HttpResponseHeaderParser.class.getName());
    
    
    
	
	private static ThreadLocal<HttpResponseHeaderParser> instanceThreadLocal = new ThreadLocal<HttpResponseHeaderParser>(); 
	

	private static final int MAX_HEADER_SIZE = 8192;
	
	
	private static final int STATE_INIT = 0;
	private static final int STATE_READING_REQUEST_LINE = 5;
	private static final int STATE_READING_HEADER_LINES = 9;
	private int state = STATE_INIT;


	private HttpResponseHeader responseHeader = null;

	private ResponseLineParser responseLineParser;
	private HeaderlineParser headerlineParser;
	
	
	/**
	 * creates a new response header parser
	 * @return a new response header parser
	 */
	static HttpResponseHeaderParser newInstance() {
		HttpResponseHeaderParser instance = instanceThreadLocal.get();
		if (instance == null) {
			instance = new HttpResponseHeaderParser();
		}
		
		instanceThreadLocal.set(null);
		
		return instance;
	}
	
	
	/**
	 * recycles a response header parser
	 * @param parser  the response header parser to recycle
	 */
	static void recycleInstance(HttpResponseHeaderParser parser) {
		instanceThreadLocal.set(parser);
	}
	
	
	/**
	 * recycle this instance
	 */
	public void recycle() {
		HttpResponseHeaderParser.recycleInstance(this);
	}
	
	
	
	/**
	 * {@inheritDoc}
	 */
	public IHttpHeader parse(INonBlockingConnection connection, ComposedByteBuffer composedByteBuffer) throws SimpleResponseMessageException, BadMessageException, IOException {
		
		int size = 0;

		switch (state) {
		
			case STATE_INIT:
				responseHeader = new HttpResponseHeader();
				responseLineParser = ResponseLineParser.newInstance();
				state = STATE_READING_REQUEST_LINE;
				return parse(connection, composedByteBuffer);
				
				
			case STATE_READING_REQUEST_LINE:
				size = computeSizeToRead(composedByteBuffer);
				for (int i = 0; i < size; i++) {
					boolean isRead = responseLineParser.parse(connection, composedByteBuffer, responseHeader);
					if (isRead) {
						responseLineParser.recycle();
						responseLineParser = null;
						headerlineParser = HeaderlineParser.newInstance();
						
						state = STATE_READING_HEADER_LINES;
						return parse(connection, composedByteBuffer);
					}
				}
				break;

				
			case STATE_READING_HEADER_LINES:
				size = computeSizeToRead(composedByteBuffer);
				for (int i = 0; i < size; i++) {
					boolean isRead = headerlineParser.parse(composedByteBuffer, responseHeader);
					if (isRead) {
						headerlineParser.recycle();
						headerlineParser = null;

						HttpResponseHeader result = responseHeader;
						responseHeader = null;
						state = STATE_INIT;

						return result;
					}
				}
				break;
				
			default:
				break;
		}

		
		if (responseHeader.incCountParsedChars(size) >= MAX_HEADER_SIZE) {
			throw new BadMessageException("max header size exceeded");			
		}
	
		return null;
	}
	
	

	private int computeSizeToRead(ComposedByteBuffer composedByteBuffer) {
		int remaining = composedByteBuffer.remaining();
		
		
		// handle max header size
		int size = remaining;
		if (remaining > MAX_HEADER_SIZE - responseHeader.getCountParsedChars()) {
			size = MAX_HEADER_SIZE - responseHeader.getCountParsedChars();
		} 
		
		return size;
	}

	
	
    public void onException(IOException ioe, ComposedByteBuffer rawData) {
        
        if (LOG.isLoggable(Level.FINE)) {
            LOG.info("error occured by receiving header. Reason: " + ioe.toString());

            StringBuilder sb = new StringBuilder();
            if (responseHeader != null) {
                sb.append(responseHeader.toString());
            } 

            LOG.info("received header lines: " +  sb.toString());
        }
    }


        

		
	
	private static final class ResponseLineParser {

		
		private static ThreadLocal<ResponseLineParser> instanceThreadLocal = new ThreadLocal<ResponseLineParser>();
		
		private int state = 0;
		

		private final StringBuilder protocolSchemeBuilder = new StringBuilder(4);
		private final StringBuilder protocolVersionBuilder = new StringBuilder(3);
		private final StringBuilder statusBuilder = new StringBuilder(3);
		private final StringBuilder reasonBuilder = new StringBuilder(8);

		static ResponseLineParser newInstance() {
			ResponseLineParser instance = instanceThreadLocal.get();
			if (instance == null) {
				instance = new ResponseLineParser();
			}
			
			instanceThreadLocal.set(null);
			
			return instance;
		}
		
		static void recycleInstance(ResponseLineParser parser) {
			instanceThreadLocal.set(parser);
		}
		
		
		public boolean parse(INonBlockingConnection connection, ComposedByteBuffer composedByteBuffer, HttpResponseHeader messageHeader) throws SimpleResponseMessageException, BadMessageException, IOException {

			byte b = composedByteBuffer.getByte();

			
			// filter CR:
			// RFC 2616 (19.3 Tolerant Applications)
			// ... The line terminator for message-header fields is the sequence CRLF.
			// However, we recommend that applications, when parsing such headers,
			// recognize a single LF as a line terminator and ignore the leading CR...

			
			switch (state) {
			
  			    // wait until '/' occurs
				case 0:
					switch (b) {
						case CR:
							break;
							
						case LF:
							break;
							
						case SPACE:
							break;
							
						case HTAB:
							break; 
							
						case '/':
							String protocolScheme = protocolSchemeBuilder.toString();
							messageHeader.setProtocolSchemeSilence(protocolScheme);
							state = 5;
							break;
							
						default:
							protocolSchemeBuilder.append((char) b);
						
							// simple response?
							if (protocolSchemeBuilder.length() > 4) {
								composedByteBuffer.addFirst(protocolSchemeBuilder.toString().getBytes("ISO-8859-1"));
								reset();
								throw new SimpleResponseMessageException();
							}
							break;
					}
					break;
					
					
				// '/' occurred
				case 5:
					protocolVersionBuilder.append((char) b);
					state = 10;
					break;

					
					
				// wait until version ends
				case 10:
					switch (b) {
						case SPACE:
							messageHeader.setProtocolVersionSilence(protocolVersionBuilder.toString());
							state = 15;
							break;
							
						case HTAB:
							messageHeader.setProtocolVersionSilence(protocolVersionBuilder.toString());
							state = 15;
							break;
							
						case CR:
							break;
							
						case LF:
							throw new BadMessageException("bad request");

						default:
							protocolVersionBuilder.append((char) b);
							break;
					}
					break;

					
					
				// wait until status starts
				case 15:
					switch (b) {
						case SPACE:
							break;
							
						case HTAB:
							break;
	
						default:
							statusBuilder.append((char) b);
							state = 20;
							break;
					}
					break;
					
					
				
				// read status
				case 20:
					switch (b) {
						case SPACE:
							messageHeader.setStatus(Integer.parseInt(statusBuilder.toString()));
							state = 30;
							break;
							
						case HTAB:
							messageHeader.setStatus(Integer.parseInt(statusBuilder.toString()));
							state = 30;
							break;
							
						case CR:
							break;
							
						case LF:
							messageHeader.setStatus(Integer.parseInt(statusBuilder.toString()));
							reset();
							return true;

						default:
							statusBuilder.append((char) b);
							break;
					}
					break;


					
				// wait until reason starts
				case 30:
					switch (b) {
						case SPACE:
							break;
							
						case HTAB:
							break;
	
						default:
							reasonBuilder.append((char) b);
							state = 35;
							break;
				}
				break;
				
				
			
				// read reason
				case 35:
					switch (b) {
							
						case CR:
							break;
							
						case LF:
							messageHeader.setReason(reasonBuilder.toString());
							reset();
							return true;
	
						default:
							reasonBuilder.append((char) b);
							break;
					}
					break;


					
					
				default:
					break;
			}
			
			
			return false;
		}
	
		
				
		public void reset() {
			state = 0;
			
			protocolSchemeBuilder.setLength(0);
			protocolVersionBuilder.setLength(0);
			statusBuilder.setLength(0);
			reasonBuilder.setLength(0);
		}
		
		
		public void recycle() {
			ResponseLineParser.recycleInstance(this);
		}
	}
}

 