/*
 *  Copyright (c) xlightweb.org, 2008 - 2009. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;



import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xsocket.DataConverter;




/**
 * a multipart byte range message body parser
 *  
 * @author grro@xlightweb.org
 */
final class MultipartByteRangeMessageBodyParser extends AbstractBodyParser {
	
	private static final Logger LOG = Logger.getLogger(MultipartByteRangeMessageBodyParser.class.getName());
	

	private final IHttpHeader header;
	private String endSeparator;
	private int separatorIdx = 0;
	
	private final StringBuilder stringBuilder = new StringBuilder(); 

	
	/**
	 * constructor 
	 * 
	 * @param httpConnection   the http connection
	 * @param header           the header
	 * @throws IOException if an exception occurs
	 */
	public MultipartByteRangeMessageBodyParser(AbstractHttpConnection httpConnection, IHttpHeader header) throws IOException {
		super(BodyType.IN_MEMORY, httpConnection, header);
		
		this.header = header;
		
		String contentType = header.getContentType();
		for (String token : contentType.split(";")) {
			token = token.trim();
			if (token.toLowerCase().startsWith("boundary=")) {
				endSeparator = "--" +  token.substring("boundary=".length(), token.length()).trim() + "--";
			}
		}
	}

	
    /**
     * {@inheritDoc}
     */
    @Override
    void onException(IOException ioe, ComposedByteBuffer rawData) {
    	if (!isComplete()) {
    		register(new ProtocolException("error occured by parsing multipart byterange message body " + ioe.toString()));
    	}
    }


    /**
	 * {@inheritDoc}
	 */
	@Override
	void onDisconnect() {
		if (!isComplete()) {
			setPersistent(false);
		}
		
		super.onDisconnect();
	}



	/**
	 * {@inheritDoc}
	 */
	@Override
	void doParse(ComposedByteBuffer rawData) throws IOException {
			
		while (rawData.available() > 0) {
			
			byte b = rawData.getByte();
			
			if (b == (byte) endSeparator.charAt(separatorIdx)) {
				separatorIdx++;
				if (separatorIdx == endSeparator.length()) {
					if (LOG.isLoggable(Level.FINE)) {
						LOG.fine("body read. set complete true");
					}
					stringBuilder.append(endSeparator);
					stringBuilder.append("\r\n");
					
					super.add(DataConverter.toByteBuffer(stringBuilder.toString(), header.getCharacterEncoding()));
					setComplete();
					return;
				}
				
			} else {
				if (separatorIdx > 0) {
					stringBuilder.append(endSeparator.substring(0, separatorIdx));
					separatorIdx = 0;
				} else {
					stringBuilder.append((char) b);
				}
			}
		}
	}
}

 