/*
 *  Copyright (c) xlightweb.org, 2008 - 2009. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */

package org.xlightweb;

import java.io.Closeable;
import java.io.IOException;
import java.net.ConnectException;
import java.net.SocketTimeoutException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.logging.Level;
import java.util.logging.Logger;


import org.xlightweb.AbstractHttpConnection.IMultimodeExecutor;
import org.xsocket.DataConverter;
import org.xsocket.Execution;
import org.xsocket.ILifeCycle;
import org.xsocket.SerializedTaskQueue;
import org.xsocket.connection.IWriteCompletionHandler;
import org.xsocket.connection.IConnection.FlushMode;




/**
 * Implements a handler chain. The next handler of the chain will be called (in the registering order) by forwarding 
 * the request. See {@link IHttpExchange}. Example:
 * 
 * <pre>
 * &#064Execution(Execution.NONTHREADED)
 * class HeaderLogInterceptor implements IHttpRequestHandler {
 *     
 *    public void onRequest(final IHttpExchange exchange) throws IOException {
 *       
 *       IHttpResponseHandler respHdl = new IHttpResponseHandler() {
 *       
 *          public void onResponse(IHttpResponse response) throws IOException {
 *             System.out.println(response.getResponseHeader());
 *             exchange.send(response);
 *          }
 *          
 *          public void onException(IOException ioe) {
 *             System.out.println(ioe.toString());
 *             exchange.sendError(500);
 *          }
 *       };
 *       
 *       System.out.println(exchange.getRequest().getRequestHeader());
 *       exchange.forward(exchange.getRequest(), respHdl);
 *    }
 * }
 * 
 * 
 * RequestHandlerChain chain = new RequestHandlerChain();
 * chain.addLast(new HeaderLogInterceptor());
 * chain.addLast(new MySeviceHandler());
 * 
 * IServer server = new HttpServer(8080, chain);
 * server.start();
 * </pre>
 * 
 * 
 * @author grro@xlightweb.org
 */

@Execution(Execution.NONTHREADED)
public class RequestHandlerChain extends RequestHandlerBase implements IHttpRequestTimeoutHandler, ILifeCycle {

	private static final Logger LOG = Logger.getLogger(RequestHandlerChain.class.getName());
	

	private final List<IHttpRequestHandler> handlers = new ArrayList<IHttpRequestHandler>();
	private final List<ILifeCycle> lifeCycleChain = new ArrayList<ILifeCycle>();

	private final ArrayList<IHttpRequestHandler> requestHandlerChain = new ArrayList<IHttpRequestHandler>();
	
	private boolean isOnRequestTimeoutPathMultithreaded = false;
	private final List<IHttpRequestTimeoutHandler> requestTimeoutHandlerChain = new ArrayList<IHttpRequestTimeoutHandler>();


	private final MultimodeExecutor executor = new MultimodeExecutor();

	
	

	/**
	 * constructor 
	 * 
	 */
	public RequestHandlerChain() {
		
	}
	
	
	/**
	 * constructor 
	 * 
	 * @param handlers the initial handlers 
	 */
	public RequestHandlerChain(List<IHttpRequestHandler> handlers) {
		for (IHttpRequestHandler hdl : handlers) {
			addLast(hdl);
		}
	}

	
	
	/**
	 * add a handler to the top of the chain
	 * 
	 * @param handler the handler to add
	 */
	public void addFirst(IHttpRequestHandler handler) {
		handlers.add(0, handler);
		
		computePath();
	}
	
	
	
	
	/**
	 * add a handler to the end of the chain
	 * 
	 * @param handler the handler to add
	 */
	public void addLast(IHttpRequestHandler handler) {
		handlers.add(handler);
		
		computePath();
	}
	
	
	/**
	 * removes the handler 
	 * 
	 * @param handler the handler to remove
	 */
	public void remove(IHttpRequestHandler handler) {
		handlers.remove(handler);
		
		computePath();
	}
	
	
	
	/**
	 * removes all handler 
	 */
	public void clear() {
		handlers.clear();
		
		computePath();
	}
	
	
	/**
	 * returns all handler 
	 * 
	 * @return the handlers
	 */
	public List<IHttpRequestHandler> getHandlers() {
		return Collections.unmodifiableList(handlers);
	}
	
	
	/**
	 * returns handler info
	 * 
	 * @return handler info
	 */
	List<String> getHandlerInfo() {
		List<String> result = new ArrayList<String>();
		
		for (int i = 0; i < handlers.size(); i++) {
			RequestHandlerInfo info = HttpUtils.getHttpRequestHandlerInfo(handlers.get(i));
			result.add("[" + i + "] " + handlers.get(i).getClass().getSimpleName() + "#" + handlers.get(i).hashCode() + 
					   " (multithreaded=" + info.isRequestHandlerMultithreaded() + 
					   ", invokeOnMessageReceived=" + info.isRequestHandlerInvokeOnMessageReceived() + ")");
		}
		
		return result;
	}
	
	
	private void computePath() {
		
		lifeCycleChain.clear();
		
		requestHandlerChain.clear();
	
		requestTimeoutHandlerChain.clear();
		isOnRequestTimeoutPathMultithreaded = false;
		
	
		
		for (IHttpRequestHandler handler : handlers) {
			RequestHandlerInfo handlerInfo = HttpUtils.getHttpRequestHandlerInfo(handler);

			if (handlerInfo.isLifeCycle()) {
				lifeCycleChain.add((ILifeCycle) handler);
			}

			if (handlerInfo.isRequestTimeoutHandler()) {
				requestTimeoutHandlerChain.add((IHttpRequestTimeoutHandler) handler);
				isOnRequestTimeoutPathMultithreaded = isOnRequestTimeoutPathMultithreaded || handlerInfo.isRequestTimeoutHandlerMultithreaded();
			}
			
			
			if (handlerInfo.isRequestHandlerSynchronizedOnSession()) {
				handler = new SessionSynchronizedRequestHandler(handler);
			}
			
			if (handlerInfo.isRequestHandlerMultithreaded()) {
				handler = new MultithreadedMessageHandler(handler);
			}
				
				
			if (handlerInfo.isRequestHandlerInvokeOnMessageReceived()) {
				handler = new OnMessageHandler(handler);
			}
					
			requestHandlerChain.add(handler);
		}		
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public void onInit() {
		for (ILifeCycle lifeCycle : lifeCycleChain) {
			lifeCycle.onInit();
		}
	}
	

	/**
	 * {@inheritDoc}
	 */
	public void onDestroy() throws IOException {
		for (ILifeCycle lifeCycle : lifeCycleChain) {
			lifeCycle.onDestroy();
		}		
		
		executor.close();
	}


	
	/**
	 * {@inheritDoc}
	 */
	@InvokeOn(InvokeOn.HEADER_RECEIVED)
	public void onRequest(IHttpExchange exchange) throws IOException {
	
		if (requestHandlerChain.isEmpty()) {
			exchange.forward(exchange.getRequest(), new HttpResponseHandler(exchange));
			return;
		} 
		
		callOnRequestCallback(exchange);
	}
	 
	
	private static final class HttpResponseHandler implements IHttpResponseHandler {

		private IHttpExchange exchange = null;
		
		public HttpResponseHandler(IHttpExchange exchange) {
			this.exchange = exchange;
		}
		
		@Execution(Execution.NONTHREADED)
		public void onResponse(IHttpResponse response) throws IOException {
			exchange.send(response);
		}
		
		@Execution(Execution.NONTHREADED)
		public void onException(IOException ioe) {
			exchange.sendError(ioe);
		}
	}
	
	
	

	@SuppressWarnings("unchecked")
	private void callOnRequestCallback(IHttpExchange exchange) throws IOException {
		
		ChainExchange chainExchange = new ChainExchange(exchange, executor, exchange.getRequest(), (List<IHttpRequestHandler>) requestHandlerChain.clone(), new HttpResponseHandler(exchange));
		chainExchange.handle();
	}
	
	

	/**
	 * {@inheritDoc}
	 */
	public boolean onRequestTimeout(IHttpConnection httpConnection)throws IOException {
	
		if (requestTimeoutHandlerChain.isEmpty()) {
			if (LOG.isLoggable(Level.FINER)) {
				LOG.finer("no request timeout handler set. ignore callback");
			}
			
			return false;
		}
		
		if (isOnRequestTimeoutPathMultithreaded) {
			((AbstractHttpConnection) httpConnection).getExecutor().processMultithreaded(new OnRequestTimeoutCaller(httpConnection));
			
		} else {
			((AbstractHttpConnection) httpConnection).getExecutor().processNonthreaded(new OnRequestTimeoutCaller(httpConnection));
		}
		
		return true;
	}
	

	
	private boolean callOnRequestTimeoutCallback(IHttpConnection connection) throws IOException {
		
		for (IHttpRequestTimeoutHandler requestTimeoutHandler : requestTimeoutHandlerChain) {
			boolean result = requestTimeoutHandler.onRequestTimeout(connection);
			if (result == true) {
				return true;
			}
		}	
		
		return false;
	}
	

	
	
	private final class OnRequestTimeoutCaller implements Runnable {
		
		private IHttpConnection httpConnection = null;
		
		public OnRequestTimeoutCaller(IHttpConnection httpConnection) {
			this.httpConnection = httpConnection;
		}
		
		public void run() {
			try {
				callOnRequestTimeoutCallback(httpConnection);
			} catch (IOException ioe) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("Error occured by calling onRequestTimeout callback " + ioe.toString());
				}
			}
		}
	}
	
	
	
	private static final class MultimodeExecutor implements IMultimodeExecutor, Closeable {

		private ExecutorService workerpool = Executors.newCachedThreadPool();
		private final SerializedTaskQueue taskQueue = new SerializedTaskQueue();
				
		public void processMultithreaded(Runnable task) {
			taskQueue.performMultiThreaded(task, workerpool);
		}
		
		public void processNonthreaded(Runnable task) {
			taskQueue.performNonThreaded(task, workerpool);
		}
		
		public void close() {
			workerpool.shutdown();
		}
	}
		
	
	private static final class ChainExchange implements IHttpExchange {
		
		private IHttpExchange exchange = null;
		private List<IHttpRequestHandler> requestHandlerChain = null;

		private final IMultimodeExecutor executor;
		
		private IHttpRequest request = null;
		private IHttpResponseHandler responseHandler = null;
		private ResponseHandlerInfo responseHandlerInfo = null;
		
		
		public ChainExchange(IHttpExchange exchange, IMultimodeExecutor executor, IHttpRequest request, List<IHttpRequestHandler> requestHandlerChain, IHttpResponseHandler responseHandler) {
			this.exchange = exchange;
			this.executor = executor;
			this.request = request;
			this.requestHandlerChain = requestHandlerChain;
			this.responseHandler = responseHandler;


			if (responseHandler != null) {
				responseHandlerInfo = HttpUtils.getResponseHandlerInfo(responseHandler);
			}
		}
		
		
		
		void handle() throws IOException {
			
			// is chain ends is reached -> calling external
			if (requestHandlerChain.isEmpty()) {
				IHttpResponseHandler respHdl = new IHttpResponseHandler() {
					
					@Execution(Execution.NONTHREADED)
					public void onResponse(IHttpResponse response) throws IOException {
						send(response);
					}
					
					public void onException(IOException ioe) {
						sendError(ioe);
					}
				};
				
				exchange.forward(request, respHdl);		
				
			// get next handler of chain
			} else {
				
				final IHttpRequestHandler requestHandler = requestHandlerChain.remove(0);
						
				if ((getRequest().hasBody()) && (HttpUtils.isContentTypeFormUrlencoded(getRequest()))) {
					ChainExchange.this.request = HttpUtils.newFormEncodedRequestWrapper(request);
				
					if (!getRequest().getNonBlockingBody().isComplete()) {
							
						IBodyCompleteListener cl = new IBodyCompleteListener() {
							
							@Execution(Execution.NONTHREADED)
							public void onComplete() throws IOException {
								performOnRequest(requestHandler);
							}
						};
							
						getRequest().getNonBlockingBody().addCompleteListener(cl);
						return;
					}
				}
		
				performOnRequest(requestHandler);
			}
		}
		
		
		private IMultimodeExecutor getExecutor() { 
			return executor;
		}
		
		private void performOnRequest(IHttpRequestHandler requestHandler) throws IOException {
	
			try {
				requestHandler.onRequest(this);
			} catch (BadMessageException bme) {
				 exchange.sendError(400, bme.getMessage());
			}
		}
		
		public IHttpConnection getConnection() {
			return exchange.getConnection();
		}
	
		public IHttpRequest getRequest() {
			return request;
		}
		
		public IHttpSession getSession(boolean create) {
			return exchange.getSession(create);
		}
		
		public String encodeURL(String url) {
			return exchange.encodeURL(url);
		}

		
		public void destroy() {
			exchange.destroy();
		}

		
		public BodyDataSink send(IHttpResponseHeader header) throws IOException {
			
			if (responseHandler == null) {
				return new BodyDataSink((AbstractHttpConnection) exchange.getConnection(), getExecutor(exchange), null, "ISO-8859-1");
			}
			
			if (header.getContentLength() != -1) {
				header.removeHeader("Content-Length");
			}
			
			if ((header.getTransferEncoding() == null)) {
				header.setHeader("Transfer-Encoding", "chunked");
			}
			
			
			final NonBlockingBodyDataSource bodyDataSource = new NonBlockingBodyDataSource(BodyType.IN_MEMORY, header.getCharacterEncoding());
			
			IMessageWriter messageWriter = new MessageWriter(bodyDataSource);
			BodyDataSink bodyDataSink = new BodyDataSink((AbstractHttpConnection) exchange.getConnection(), getExecutor(exchange), messageWriter, header.getCharacterEncoding());
			
			send(new HttpResponse(header, bodyDataSource));
			
			return bodyDataSink;
		}

		
		public BodyDataSink send(IHttpResponseHeader header, int contentLength) throws IOException {
			
			if (responseHandler == null) {
				return new BodyDataSink((AbstractHttpConnection) exchange.getConnection(), getExecutor(exchange), null, "ISO-8859-1");
			}
			
			if ((header.getTransferEncoding() != null) && (header.getTransferEncoding().equalsIgnoreCase("chunked"))) {
				header.removeHeader("Transfer-Encoding");
			}
			
			if (header.getContentLength() == -1) {
				header.setContentLength(contentLength);
			}
			
			
			final NonBlockingBodyDataSource bodyDataSource = new NonBlockingBodyDataSource(BodyType.IN_MEMORY, header.getCharacterEncoding());
			
			IMessageWriter messageWriter = new MessageWriter(bodyDataSource);
			BodyDataSink bodyDataSink = new BodyDataSink((AbstractHttpConnection) exchange.getConnection(), getExecutor(exchange), messageWriter, header.getCharacterEncoding());
			
			send(new HttpResponse(header, bodyDataSource));
			
			return bodyDataSink;
		}

		
		public void send(final IHttpResponse response) throws IOException {

			if (responseHandler == null) {
				LOG.warning("response will not been send, because no response handler is assigned");
				return;
			}
			
			
			if (responseHandlerInfo.isResponseHandlerInvokeOnMessageReceived()) {
				
				IBodyCompleteListener copmleteListener = new IBodyCompleteListener() {
					
					@Execution(Execution.NONTHREADED)
					public void onComplete() throws IOException {
						performOnResponse(response);
					}
				};
				response.getNonBlockingBody().addCompleteListener(copmleteListener);
				
			} else {
				performOnResponse(response);
			}
		}

		
		private void performOnResponse(final IHttpResponse response) throws IOException {

			
			Runnable task = new Runnable() {
				
				public void run() {
					try {
						responseHandler.onResponse(response);
					} catch (IOException ioe) {
						if (LOG.isLoggable(Level.FINE)) {
							LOG.fine("error occured by performing on response on " + responseHandler);
						}
					}
				}
			};

			
			if (responseHandlerInfo.isResponseHandlerMultithreaded()) {
				performMultithreaded(task);
			} else {
				performNonthreaded(task);
			}
		}
		
		
		
		public void sendError(final Exception e) throws IllegalStateException {
			
			if (responseHandlerInfo.isSocketTimeoutHandler() && (e instanceof SocketTimeoutException)) {

				Runnable task = new Runnable() {
					public void run() {
						((IHttpSocketTimeoutHandler) responseHandler).onException((SocketTimeoutException) e);
					}
				};
				
				if (responseHandlerInfo.isSocketTimeoutHandlerMultithreaded()) {
					performMultithreaded(task);
				} else {
					performNonthreaded(task);
				}
				
				
			} else if (e instanceof IOException) {
				
				Runnable task = new Runnable() {
					public void run() {
						try {
							responseHandler.onException((IOException) e);
						} catch (IOException ioe) {
							if (LOG.isLoggable(Level.FINE)) {
								LOG.fine("Error occured by performing onException " + responseHandler + ". reason: " + ioe.toString());
							}
						}
					}
				};
				
				
				if (responseHandlerInfo.isResponseExeptionHandlerMultithreaded()) {
					performMultithreaded(task);
				} else {
					performNonthreaded(task);
				}

				
			} else {
				if (HttpUtils.isShowDetailedError()) {
					sendError(500, DataConverter.toString(e));
				} else {
					sendError(500);
				}
			}
		}
		
		
		public void sendError(int errorCode) {
			sendError(errorCode, HttpUtils.getReason(errorCode));			
		}
		
		
		public void sendError(int errorCode, String msg) {
			
			try {
				
				String id = "<unset>";
				if (exchange.getConnection() != null) {
					id = exchange.getConnection().getId();
				}
				
				send(new HttpResponse(errorCode, "text/html", AbstractHttpConnection.generateErrorMessageHtml(errorCode, msg, id)));
			} catch (IOException ioe) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("could not send error message " + errorCode + " reason " + ioe.toString());
				}
				destroy();
			}
		}
		

		
		public BodyDataSink forward(IHttpRequestHeader requestHeader) throws IOException, ConnectException {
			return forward(requestHeader, new ForwardingResponseHandler(this));
		}
		
		public BodyDataSink forward(IHttpRequestHeader requestHeader, IHttpResponseHandler responseHandler) throws IOException, ConnectException {
			
			if (responseHandler == null) {
				 responseHandler = new DoNothingResponseHandler();
			 }
			
			if (requestHeader.getContentLength() != -1) {
				requestHeader.removeHeader("Content-Length");
			}
			
			if (requestHeader.getMethod().equalsIgnoreCase("GET") || requestHeader.getMethod().equalsIgnoreCase("HEAD")) {
				throw new IOException(requestHeader.getMethod() + " is a bodyless request");
			}

			if ((requestHeader.getTransferEncoding() == null)) {
				requestHeader.setHeader("Transfer-Encoding", "chunked");
			}
			

			final NonBlockingBodyDataSource bodyDataSource = new NonBlockingBodyDataSource(BodyType.IN_MEMORY, requestHeader.getCharacterEncoding());
			HttpRequest request = new HttpRequest(requestHeader, bodyDataSource);
			
			IMessageWriter messageWriter = new MessageWriter(bodyDataSource);
			BodyDataSink bodyDataSink = new BodyDataSink((AbstractHttpConnection) exchange.getConnection(), getExecutor(exchange), messageWriter, requestHeader.getCharacterEncoding());
			
			ChainExchange chainExchange = new ChainExchange(exchange, executor, request, requestHandlerChain, responseHandler);
			chainExchange.handle();
			
			return bodyDataSink;
		}
		
		
		public BodyDataSink forward(IHttpRequestHeader requestHeader, int contentLength) throws IOException, ConnectException {
			return forward(requestHeader, contentLength, new ForwardingResponseHandler(this));
		}
		
		
		public BodyDataSink forward(IHttpRequestHeader requestHeader, int contentLength, IHttpResponseHandler responseHandler) throws IOException, ConnectException {
			
			if (responseHandler == null) {
				 responseHandler = new DoNothingResponseHandler();
			 }
			
			if ((requestHeader.getTransferEncoding() != null) && (requestHeader.getTransferEncoding().equalsIgnoreCase("chunked"))) {
				requestHeader.removeHeader("Transfer-Encoding");
			}
			
			if (requestHeader.getMethod().equalsIgnoreCase("GET") || requestHeader.getMethod().equalsIgnoreCase("HEAD")) {
				throw new IOException(requestHeader.getMethod() + " is a bodyless request");
			}

			if (requestHeader.getContentLength() == -1) {
				requestHeader.setContentLength(contentLength);
			}


			NonBlockingBodyDataSource bodyDataSource = new NonBlockingBodyDataSource(BodyType.IN_MEMORY, requestHeader.getCharacterEncoding());
			HttpRequest request = new HttpRequest(requestHeader, bodyDataSource);
			
			IMessageWriter messageWriter = new MessageWriter(bodyDataSource);
			BodyDataSink bodyDataSink = new BodyDataSink((AbstractHttpConnection) exchange.getConnection(), getExecutor(exchange), messageWriter, requestHeader.getCharacterEncoding());
			
			ChainExchange chainExchange = new ChainExchange(exchange, executor, request, requestHandlerChain, responseHandler);
			chainExchange.handle();
			
			return bodyDataSink;
		}
		
		
		public void forward(IHttpRequest request) throws IOException, ConnectException {			
			forward(request, new ForwardingResponseHandler(this));
		}
		
		
		public void forward(IHttpRequest request, IHttpResponseHandler responseHandler) throws IOException, ConnectException {
			if (responseHandler == null) {
				 responseHandler = new DoNothingResponseHandler();
			 }
			
			ChainExchange chainExchange = new ChainExchange(exchange, executor, request, requestHandlerChain, responseHandler);
			chainExchange.handle();
		}
		
		
		private void performMultithreaded(Runnable task) {
			AbstractHttpConnection con = (AbstractHttpConnection) exchange.getConnection();
			if (con != null) {
				con.getExecutor().processMultithreaded(task);
			} else {
				executor.processMultithreaded(task);
			}
		}
		
		private void performNonthreaded(Runnable task) {
			AbstractHttpConnection con = (AbstractHttpConnection) exchange.getConnection();
			if (con != null) {
				con.getExecutor().processNonthreaded(task);
			} else {
				task.run();
			}
		}
		
		
		IMultimodeExecutor getExecutor(IHttpExchange exchange) {
			AbstractHttpConnection con = (AbstractHttpConnection) exchange.getConnection();
			if (con != null) {
				return con.getExecutor();
			} else {
				return executor;
			}
		}
	}
	
	
	
	private static final class MessageWriter implements IMessageWriter {
		 
		private NonBlockingBodyDataSource bodyDataSource = null;
		
		public MessageWriter(NonBlockingBodyDataSource bodyDataSource) {
			this.bodyDataSource = bodyDataSource;
		}
				 
		public void flush(ByteBuffer[] bodyData, boolean isContentImmutable, FlushMode flushMode, IWriteCompletionHandler completionHandler)throws IOException {
		    bodyDataSource.append(isContentImmutable, bodyData, completionHandler);
		}
		
		public void close() throws IOException {
			bodyDataSource.setComplete(true);
		}
		
		public void destroy() {
			bodyDataSource.destroy();
		}
		
		public boolean isNetworkEndpoint() {
			return false;
		}
		
		public int getPendingWriteDataSize() {
			return 0;
		}
	};
	

		
	
	@Execution(Execution.NONTHREADED)
	private static final class OnMessageHandler implements IHttpRequestHandler {
		
		private IHttpRequestHandler delegee = null;
		
		public OnMessageHandler(IHttpRequestHandler delegee) {
			this.delegee = delegee;
		}
		
		
		public void onRequest(final IHttpExchange exchange) throws IOException {
			
			if (exchange.getRequest().hasBody()) {
				NonBlockingBodyDataSource bodyDataSource = exchange.getRequest().getNonBlockingBody();
			
				if (!bodyDataSource.isComplete()) {
					IBodyCompleteListener cl = new IBodyCompleteListener() {
						
						@Execution(Execution.NONTHREADED)
						public void onComplete() throws IOException {
							try {
								delegee.onRequest(exchange);
							} catch (BadMessageException bme) {
								exchange.sendError(bme);
							}
						}
					};
					bodyDataSource.addCompleteListener(cl);
					
					return;
				}
				
			} 
		
			try {
				delegee.onRequest(exchange);
			 } catch (BadMessageException bme) {
				 exchange.sendError(400, bme.getMessage());
			 }
		}
		
	}
	

	private static final class SessionSynchronizedRequestHandler implements IHttpRequestHandler {
		
		private IHttpRequestHandler delegee = null;
		
		public SessionSynchronizedRequestHandler(IHttpRequestHandler delegee) {
			this.delegee = delegee;
		}
		
		
		public void onRequest(IHttpExchange exchange) throws IOException {
			
			IHttpSession session = exchange.getSession(false);
			
			if (session == null) {
				delegee.onRequest(exchange);
				
			} else {
				synchronized (session) {
					delegee.onRequest(exchange);
				}
			}
		}
	}
	
	
	
	private static final class MultithreadedMessageHandler implements IHttpRequestHandler {
		
		private IHttpRequestHandler delegee = null;
		
		public MultithreadedMessageHandler(IHttpRequestHandler delegee) {
			this.delegee = delegee;
		}
		
		
		public void onRequest(final IHttpExchange exchange) throws IOException {
			
			Runnable task = new Runnable() {
				
				public void run() {
					try {
						delegee.onRequest(exchange);
						
					} catch (BadMessageException bme) {
						exchange.sendError(bme);
						 
					} catch (IOException ioe) {
						if (LOG.isLoggable(Level.FINE)) {
							LOG.fine("error occured by performing onRequest call back " + ioe.toString());
						}
					}
				}
				
			};
			
			AbstractHttpConnection con = (AbstractHttpConnection) exchange.getConnection();
			if (con != null) {
				con.getExecutor().processMultithreaded(task);
			} else {
				((ChainExchange) exchange).getExecutor().processMultithreaded(task);
			}
		}
		
		@Override
		public String toString() {
			return super.toString() + " (delegee: " + delegee.toString() + ")";
		}
	}

	
	@Execution(Execution.NONTHREADED)
	private static final class ForwardingResponseHandler implements IHttpResponseHandler {

		private IHttpExchange exchange = null;
		
		public ForwardingResponseHandler(IHttpExchange exchange) {
			this.exchange = exchange;
		}

		
		public void onResponse(IHttpResponse response) throws IOException {
			exchange.send(response);
		}

		public void onException(IOException ioe) throws IOException {
			exchange.sendError(ioe);
		}
	}
	
	@Execution(Execution.NONTHREADED)
	static final class DoNothingResponseHandler implements IHttpResponseHandler {

		public void onResponse(IHttpResponse response) throws IOException { }

		public void onException(IOException ioe) throws IOException { }
	}
}
