/*
 *  Copyright (c) xlightweb.org, 2008 - 2009. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xsocket.DataConverter;
import org.xsocket.connection.IConnection.FlushMode;
import org.xsocket.connection.IWriteCompletionHandler;




/**
 * Simple message writer 
 *  
 * @author grro@xlightweb.org
 */
class SimpleMessageWriter extends AbstractNetworkMessageWriter {

	private static final Logger LOG = Logger.getLogger(SimpleMessageWriter.class.getName());
	
	private boolean isHeaderWritten = false;
	private final IHttpHeader header;

	
	/**
	 * constructor
	 *  
	 * @param httpConnection   the http connection
	 * @param header           the header
	 * @throws IOException if an exception occurs
	 */
	public SimpleMessageWriter(AbstractHttpConnection httpConnection, boolean destroyOnClose, IHttpHeader header) throws IOException {
		super(httpConnection, destroyOnClose);
		this.header = header;
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final void flush(ByteBuffer[] bodyData, boolean isContentImmutable, FlushMode flushMode, IWriteCompletionHandler completionHandler) throws IOException {
		AbstractHttpConnection con = getConnection();
		
		if (con != null) {
			
			try {
				con.setFlushmode(flushMode);
				
				int dataWritten = 0;
				
				if (!isHeaderWritten) {
					isHeaderWritten = true;
					
					dataWritten = con.write(header.toString() + "\r\n");			
					dataWritten += writeBody(con, bodyData, completionHandler);
					
					con.incCountMessageSent();
					
				} else {
					dataWritten = writeBody(con, bodyData, completionHandler);
				}
		
				if (dataWritten > 0) {
					con.flush();
				}
				
			} catch (IOException ioe) {
				if (LOG.isLoggable(Level.FINE)) {
					LOG.fine("[" + getId()+ "] error occured by flushing bound data sink. Destroying connection. reason " + DataConverter.toString(ioe));
				}
				
				destroy();
				throw ioe;
			}
		}
	}
	

	/**
	 * write body data
	 *  
	 * @param con                 the http connection
	 * @param bodyData            the data to write
	 * @param completionHandler   the completion handler
	 * @return number of written bytes 
	 * @throws IOException if an exception occurs
	 */
	protected int writeBody(AbstractHttpConnection con, ByteBuffer[] bodyData, IWriteCompletionHandler completionHandler) throws IOException {
		if (bodyData != null) {
		    return (int) con.write(bodyData, completionHandler);
		} else {
			return 0;
		}
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	void onClose(boolean destroyOnClose) throws IOException {
		if (getConnection() == null) {
			return;
		}
		
		destroy();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public final boolean isNetworkEndpoint() {
		return true;
	}
}