/*
 *  Copyright (c) xlightweb.org, 2008 - 2009. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xlightweb.AbstractHttpConnection.IMultimodeExecutor;
import org.xlightweb.HttpUtils.CompletionHandlerInfo;
import org.xsocket.connection.IWriteCompletionHandler;




/**
 * 
 * 
 * @author grro@xlightweb.org
 *
 */
final class WriteCompletionManager {

    private static final Logger LOG = Logger.getLogger(WriteCompletionManager.class.getName());
    
    private final String connectionId;

    
    public WriteCompletionManager(String connectionId) {
        this.connectionId = connectionId;
    }
    
    
    // write completion handler support
    private final Map<WriteCompletionHolder, List<ByteBuffer>> pendingCompletionConfirmations = new HashMap<WriteCompletionHolder, List<ByteBuffer>>();
    private AtomicBoolean isWriteCompletionSupportActivated = new AtomicBoolean(false);

    
    
    public boolean isPendingCompletionConfirmationsEmtpy() {
        synchronized (this) {
            return pendingCompletionConfirmations.isEmpty();
        }
    }
    
    
    public void registerCompletionHandler(IWriteCompletionHandler writeCompletionHandler, IMultimodeExecutor executor, ByteBuffer[] buffersToWrite) {
        
        WriteCompletionHolder holder = new WriteCompletionHolder(writeCompletionHandler, executor, buffersToWrite, connectionId);
        
        synchronized (this) {
            isWriteCompletionSupportActivated.set(true);
            pendingCompletionConfirmations.put(holder, new ArrayList<ByteBuffer>(Arrays.asList(buffersToWrite)));
        }
        
        if (LOG.isLoggable(Level.FINE)) {
            int size = 0;
            for (ByteBuffer byteBuffer : buffersToWrite) {
                size += byteBuffer.remaining();
            }
            LOG.fine("[" + connectionId + "] {" + buffersToWrite.hashCode() + "} registering " + writeCompletionHandler.getClass().getSimpleName() + "#" + writeCompletionHandler.hashCode() + " for  data  size=" + size + " bytes)");
        }
    }
    
    

    public void onWritten(ByteBuffer[] data, boolean isForceMultithreaded) {
    	if (isWriteCompletionSupportActivated.get()) {
    		for (ByteBuffer byteBuffer : data) {
    			handleOnWritten(byteBuffer, isForceMultithreaded);
    		} 
        }
    }   
    
    public void onWritten(ByteBuffer data, boolean isForceMultithreaded) {
    	if (isWriteCompletionSupportActivated.get()) {
    		handleOnWritten(data, isForceMultithreaded);
        }
    }   
    
    
    public List<ByteBuffer> getPendungBuffers() {
    	List<ByteBuffer> buffers = new ArrayList<ByteBuffer>();
    	
    	for (List<ByteBuffer> byteBuffers : pendingCompletionConfirmations.values()) {
			buffers.addAll(byteBuffers);
		}
    	 
    	return  buffers;
    }
    
    
    
    private void handleOnWritten(ByteBuffer data, boolean isForceMultithreaded) {

        WriteCompletionHolder holderToExecute = null;
        
        if (data != null) {
        	
        	synchronized (this) {
        		for (Entry<WriteCompletionHolder, List<ByteBuffer>> entry : pendingCompletionConfirmations.entrySet()) {
	                List<ByteBuffer> buffers = entry.getValue();
	                for (ByteBuffer buffer : buffers) {
	                    if (buffer == data) {
	                        buffers.remove(data);
	                        break;
	                    }
	                }
	                
	                if (buffers.isEmpty()) {
	                    pendingCompletionConfirmations.remove(entry.getKey());
	                    holderToExecute = entry.getKey();
	                    break;
	                }
        		}
            }
        }
        
        if (holderToExecute != null) {
            holderToExecute.performOnWritten(isForceMultithreaded);
        }
    }

        

    public void onWriteException(IOException ioException, ByteBuffer[] data) {

    	if (isWriteCompletionSupportActivated.get()) {
    		for (ByteBuffer byteBuffer : data) {
    			onWriteException(ioException, byteBuffer);
            }
        }
    }
        

        
    private void onWriteException(IOException ioException, ByteBuffer data) {

        WriteCompletionHolder holderToExecute = null;
        
        if (data != null) {
        	synchronized (this) {
	            outer : for (Entry<WriteCompletionHolder, List<ByteBuffer>> entry : pendingCompletionConfirmations.entrySet()) {
	                List<ByteBuffer> buffers = entry.getValue();
	                for (ByteBuffer buffer : buffers) {
	                    if (buffer == data) {
	                        buffers.remove(data);
	                        holderToExecute = entry.getKey();
	                        break outer;
	                    }
	                }
	            }
        	}
        }
        
        if (holderToExecute != null) {
            holderToExecute.performOnException(ioException);
        }
    }
    
    
    public static final class WriteCompletionHolder implements Runnable {
        
        private final IWriteCompletionHandler handler;
        private final CompletionHandlerInfo handlerInfo;
        private final IMultimodeExecutor executor;
        private final String id;
        private final ByteBuffer[] bufs;
        private final int size;

        

        public WriteCompletionHolder(IWriteCompletionHandler handler, IMultimodeExecutor executor, ByteBuffer[] bufs, String id) {
            this.handler = handler;
            this.executor = executor;
            this.handlerInfo = HttpUtils.getCompletionHandlerInfo(handler);
            this.id = id;
            this.bufs = bufs;
            this.size = computeSize(bufs);
        }

        
        private static int computeSize(ByteBuffer[] bufs) {
        	int i = 0;
        	for (ByteBuffer byteBuffer : bufs) {
        		i += byteBuffer.remaining();
            }
        	
        	return i;
        }
        
      
        void performOnWritten(boolean isForceMultithreaded) {
        	
            if (isForceMultithreaded || handlerInfo.isOnWrittenMultithreaded()) {
                executor.processMultithreaded(this);
              
            } else {
                executor.processNonthreaded(this);
            }
        }
      
      
        public void run() {
            callOnWritten();
        }

        private void callOnWritten() {
            
            try {
                if (LOG.isLoggable(Level.FINE)) {
                    LOG.fine("[" + id + "] {" + bufs.hashCode() + "}  data (size=" + size + " bytes) has been written. calling " + handler.getClass().getSimpleName() + "#" + handler.hashCode() +  " onWritten method");
                }
                handler.onWritten(size);
            } catch (IOException ioe) {
                if (LOG.isLoggable(Level.FINE)) {
                    LOG.fine("[" + id + "] {" + bufs.hashCode() + "}   error occured by calling onWritten " + ioe.toString() + " closing connection");
                }
                
                performOnException(ioe);
            }
        }

      
        void performOnException(final IOException ioe) {
            if (handlerInfo.isOnExceptionMutlithreaded()) {
                Runnable task = new Runnable() {
                    public void run() {
                        callOnException(ioe);
                    }
                };
                executor.processMultithreaded(task);
              
            } else {
                Runnable task = new Runnable() {
                    public void run() {
                        callOnException(ioe);
                    }
                };
                executor.processNonthreaded(task);
            }
        }
      
      
        private void callOnException(IOException ioe) {
            if (LOG.isLoggable(Level.FINE)) {
                LOG.fine("[" + id + "] calling " + handler.getClass().getSimpleName() + "#" + handler.hashCode() +  " onException with " + ioe.toString());
            }

            handler.onException(ioe);
        }
    } 
}
